"""
    CircleGeometry.py
"""
import numpy as np

"""
    borrowed @
    https://stackoverflow.com/questions/55816902/finding-the-intersection-of-two-circles 
"""
def get_intersections(x0, y0, r0, x1, y1, r1):
    # circle 1: (x0, y0), radius r0
    # circle 2: (x1, y1), radius r1

    d=np.sqrt((x1-x0)**2 + (y1-y0)**2)

    # non intersecting
    if d > r0 + r1 :
        return None
    # One circle within other
    if d < abs(r0-r1):
        return None
    # coincident circles
    if d == 0 and r0 == r1:
        return None
    else:
        a=(r0**2-r1**2+d**2)/(2*d)
        h=np.sqrt(r0**2-a**2)
        x2=x0+a*(x1-x0)/d   
        y2=y0+a*(y1-y0)/d   
        x3=x2+h*(y1-y0)/d     
        y3=y2-h*(x1-x0)/d 
        x4=x2-h*(y1-y0)/d
        y4=y2+h*(x1-x0)/d
        return np.array([x3, y3]), np.array([x4, y4])

"""
    borrowed @
    https://www.johndcook.com/blog/2023/08/27/intersect-circles/
"""
def circle_intersect(x0, y0, r0, x1, y1, r1):
    c0 = np.array([x0, y0])
    c1 = np.array([x1, y1])
    v = c1 - c0
    d = np.linalg.norm(v)

    if d > r0 + r1 or d == 0:
        return None
    
    u = v/np.linalg.norm(v)
    xvec = c0 + (d**2 - r1**2 + r0**2)*u/(2*d)

    uperp = np.array([u[1], -u[0]])
    a = ((-d+r1-r0)*(-d-r1+r0)*(-d+r1+r0)*(d+r1+r0))**0.5/d
    return (xvec + a*uperp/2, xvec - a*uperp/2)

"""
    borrowed @
    https://stackoverflow.com/questions/30844482/what-is-most-efficient-way-to-find-the-intersection-of-a-line-and-a-circle-in-py
"""
def circle_line_segment_intersection(circle_center, circle_radius, pt1, pt2, full_line=True, tangent_tol=1e-9):
    """ Find the points at which a circle intersects a line-segment.  This can happen at 0, 1, or 2 points.

    :param circle_center: The (x, y) location of the circle center
    :param circle_radius: The radius of the circle
    :param pt1: The (x, y) location of the first point of the segment
    :param pt2: The (x, y) location of the second point of the segment
    :param full_line: True to find intersections along full line - not just in the segment.  False will just return intersections within the segment.
    :param tangent_tol: Numerical tolerance at which we decide the intersections are close enough to consider it a tangent
    :return Sequence[Tuple[float, float]]: A list of length 0, 1, or 2, where each element is a point at which the circle intercepts a line segment.

    Note: We follow: http://mathworld.wolfram.com/Circle-LineIntersection.html
    """

    (p1x, p1y), (p2x, p2y), (cx, cy) = pt1, pt2, circle_center
    (x1, y1), (x2, y2) = (p1x - cx, p1y - cy), (p2x - cx, p2y - cy)
    dx, dy = (x2 - x1), (y2 - y1)
    dr = (dx ** 2 + dy ** 2)**.5
    big_d = x1 * y2 - x2 * y1
    discriminant = circle_radius ** 2 * dr ** 2 - big_d ** 2

    if discriminant < 0:  # No intersection between circle and line
        return []
    else:  # There may be 0, 1, or 2 intersections with the segment
        intersections = [
            (cx + (big_d * dy + sign * (-1 if dy < 0 else 1) * dx * discriminant**.5) / dr ** 2,
             cy + (-big_d * dx + sign * abs(dy) * discriminant**.5) / dr ** 2)
            for sign in ((1, -1) if dy < 0 else (-1, 1))]  # This makes sure the order along the segment is correct
        if not full_line:  # If only considering the segment, filter out intersections that do not fall within the segment
            fraction_along_segment = [(xi - p1x) / dx if abs(dx) > abs(dy) else (yi - p1y) / dy for xi, yi in intersections]
            intersections = [pt for pt, frac in zip(intersections, fraction_along_segment) if 0 <= frac <= 1]
        if len(intersections) == 2 and abs(discriminant) <= tangent_tol:  # If line is tangent to circle, return just one point (as both intersections have same location)
            return [intersections[0]]
        else:
            return intersections


"""
    SimpleSecParams.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.SecTheory.RetensionTime import (make_initial_guess, compute_retention_time,
                                     estimate_conformance_params,
                                     estimate_conformance_params_fixed_poreexponent,
                                    )
from molass_legacy.SecTheory.ColumnConstants import SECCONF_LOWER_BOUND, BAD_CONFORMANCE_REDUCE

SEC_PENALTY_SCALE = 1e8
MIN_RETENTION_TIME_RATIO = 0.2

def initial_guess(xr_params):
    params, bounds = make_initial_guess(xr_params[:,1])
    """
    note: this won't improve bounds for rp, m in the succession of "Known Best"
    """
    return Struct(params=params, bounds=bounds)

def sec_comformance(xr_params, rg_params, seccol_params, poresize_bounds=None):
    # model_trs = compute_retention_time(seccol_params, rg_params)
    # task: unify compute_retention_time and the code below

    Npc, rp, tI, t0, P, m = seccol_params
    rhos = rg_params/rp
    rhos[rhos > 1] = 1
    model_trs = tI + P*(1 - rhos)**m

    if poresize_bounds is None:
        poresize_penalty = 0
    else:
        poresize_penalty = min(0, rp - poresize_bounds[0])**2 + max(0, rp - poresize_bounds[1])**2

    min_t = P*MIN_RETENTION_TIME_RATIO
    other_penalty = min(0, t0 - tI - min_t)**2 + min(0, m - 1)**2 + max(0, m - 3)**2
    log_conformance = np.log10(np.average((model_trs - xr_params[:,1])**2) + SEC_PENALTY_SCALE*(poresize_penalty + other_penalty))
    if log_conformance > 0:
        log_conformance *= BAD_CONFORMANCE_REDUCE   # large conformance at early stages can be misleading

    return max(SECCONF_LOWER_BOUND, log_conformance)

class SimpleSecParams:
    def __init__(self, poresize, poreexponent):
        self.poresize = poresize
        self.poreexponent = poreexponent

        if poreexponent is None or poreexponent == 0:   # poreexponent == 0 is used since set_setting("poreexponent", None) doesn't seem to work
            self.init_method = initial_guess
            self.conf_method = sec_comformance
            self.estm_method = estimate_conformance_params
            self.nump_adjust = 0
        else:
            self.init_method = self.initial_guess_fixed_poreexponent
            self.conf_method = self.sec_comformance_fixed_poreexponent
            self.estm_method = self.estimate_conformance_params_fixed_poreexponent
            self.nump_adjust = -1

    def initial_guess_fixed_poreexponent(self, xr_params):
        guess = initial_guess(xr_params)
        return Struct(params=guess.params[0:3], bounds=guess.bounds[0:3])

    def sec_comformance_fixed_poreexponent(self, xr_params, rg_params, seccol_params):
        return sec_comformance(xr_params, rg_params, np.concatenate([seccol_params, [self.poreexponent]]))

    def estimate_conformance_params_fixed_poreexponent(self, rgs, trs):
        result = estimate_conformance_params_fixed_poreexponent(rgs, trs, self.poreexponent)
        result.x = np.concatenate([result.x, [self.poreexponent]])
        return result



"""
    DenssGui.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import sys
import os
import re
import logging
from time import sleep
import numpy as np
import queue
import matplotlib.pyplot    as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.KillableThread import Thread
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, ScrolledText, is_empty_val
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.TkCustomWidgets import FileEntry, FolderEntry
from molass_legacy.KekLib.ReadOnlyText import ReadOnlyText
import molass_legacy.KekLib.CustomMessageBox as MessageBox
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.OurMatplotlib import get_color
from .DenssUtils import get_outfolder
from molass_legacy.KekLib.NumpyUtils import np_loadtxt
import molass_legacy.KekLib.DebugPlot as dplt

class StdoutRedirector:
    def __init__(self, queue):
        self.saved_stdout = sys.stdout
        self.queue = queue
        sys.stdout = self

    def write(self, string):
        self.queue.put([1, string])

    def flush(self):
        pass

    def __del__(self):
        sys.stdout = self.saved_stdout
        print('__del__ ok')

class ProgressLoghandler(logging.Handler):
    def __init__(self, log_text, log_file):
        logging.Handler.__init__(self)
        self.log_text   = log_text
        self.fh = open(log_file, 'w')
        self.formatter  = logging.Formatter('%(asctime)s %(message)s', '%H:%M:%S')

    def emit( self, record ):
        if record.levelno >= logging.INFO:
            try:
                log_message = self.formatter.format(record) + '\n'
                self.fh.write(log_message)
                self.fh.flush()
                self.log_text.insert(Tk.INSERT, log_message)
                self.log_text.see(Tk.END)
            except:
                # this error can occur when canceled
                pass

    def __del__(self):
        self.fh.close()

log_step_re = re.compile(r'^\s*(\d+)')
log_maxnum_re = re.compile(r'Maximum number of steps: (\d+)')

class DenssGuiDialog(Dialog):
    def __init__(self, parent, q=None, a=None, e=None, infile_name=None, debug=False):
        self.debug = debug
        if debug:
            import logging
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = None
        self.grab = 'local'     # used in grab_set
        self.parent = parent
        self.denss_action = False   # used for test control in test_990_DenssGui.py
        self.q = q
        self.a = a
        self.e = e
        self.dmax = Tk.DoubleVar()
        self.dmax.set(100)
        if q is None:
            # as invoked from the DENSS tools menu
            pass
        else:
            # as invoked from the LRF preview
            self.prepare_from_args(q, a, e, infile_name)
        self.cwd_init = os.getcwd()
        self.thread = None
        from .denss._version import __version__ as denss_version
        Dialog.__init__( self, parent, "Gui for DENSS-%s" % denss_version, visible=False )

    def show(self):
        self._show()

    def add_dnd_bind(self):
        self.mpl_canvas_widget.register_drop_target("*")

        def dnd_handler(event):
            self.on_fig_dnd(event)

        self.mpl_canvas_widget.bind("<<Drop>>", dnd_handler)

    def body(self, body_frame):
        set_icon( self )

        upper_frame = Tk.Frame(body_frame)
        upper_frame.pack()
        lower_frame = Tk.Frame(body_frame)
        lower_frame.pack(fill=Tk.X)

        cframe = Tk.Frame(upper_frame)
        cframe.pack(side=Tk.LEFT, padx=10, pady=5)

        iframe = Tk.Frame(upper_frame)
        iframe.pack(side=Tk.LEFT, padx=10, pady=5)

        self.create_figure(cframe)

        grid_row = 0
        io_frame = Tk.Frame(iframe)
        io_frame.grid(row=grid_row, column=0)

        grid_row_io = 0

        if self.q is None:
            self.is_memory_input = False
            in_file_label = Tk.Label(io_frame, text='Input File: ')
            in_file_label.grid(row=grid_row_io, column=0, pady=5)
            self.in_file = Tk.StringVar()
            self.in_file.set(None)
            self.in_file_entry = FileEntry(io_frame, textvariable=self.in_file, width=70, on_entry_cb=self.on_file_entry)
            self.in_file_entry.grid(row=grid_row_io, column=1, pady=10)
        else:
            self.is_memory_input = True
            self.input_info_label = Tk.Label(io_frame, text='Input is from the memory equivalent to %s' % self.infile_name, bg='white')
            self.input_info_label.grid(row=grid_row_io, column=0, columnspan=2, pady=10)

        grid_row_io += 1
        dmax_label = Tk.Label(io_frame, text='Dmax: ')
        dmax_label.grid(row=grid_row_io, column=0, sticky=Tk.E)
        dmax_entry_frame = Tk.Frame(io_frame)
        dmax_entry_frame.grid(row=grid_row_io, column=1, columnspan=2, sticky=Tk.W)
        self.dmax_entry = Tk.Entry(dmax_entry_frame, textvariable=self.dmax, width=8, justify=Tk.CENTER)
        self.dmax_entry.grid(row=0, column=0, sticky=Tk.W)
        self.illust_btn = Tk.Button(dmax_entry_frame, text="Illustrate", command=self.illustrate_dmax)
        self.illust_btn.grid(row=0, column=1, padx=40)
        self.fit_data_btn = Tk.Button(dmax_entry_frame, text="Show denss.fit_data.py figure", command=self.denss_fit_data)
        self.fit_data_btn.grid(row=0, column=12)

        grid_row_io += 1
        out_folder_label = Tk.Label(io_frame, text='Output Folder: ')
        out_folder_label.grid(row=grid_row_io, column=0)

        out_folder_init = get_outfolder(parent=self.parent)

        self.out_folder = Tk.StringVar()
        self.out_folder.set(out_folder_init)
        self.out_folder_entry = FolderEntry(io_frame, textvariable=self.out_folder, width=70, on_entry_cb=self.on_folder_entry)
        self.out_folder_entry.grid(row=grid_row_io, column=1, sticky=Tk.W, pady=5)

        grid_row += 1
        self.progress_label = Tk.Label(iframe, text="DENSS Progress" )
        self.progress_label.grid(row=grid_row, column=0)

        grid_row += 1
        self.progress_text = ReadOnlyText(iframe, width=80, height=5)
        self.progress_text.grid(row=grid_row, column=0)
        self.progress_text.update()

        grid_row += 1
        length = self.progress_text.winfo_width()
        self.progress_bar = ttk.Progressbar(iframe, orient ="horizontal", length=length, mode="determinate")
        self.progress_bar.grid(row=grid_row, column=0, pady=10)

        grid_row += 1
        self.log_text_label = Tk.Label(iframe, text="Log from DENSS" )
        self.log_text_label.grid(row=grid_row, column=0)

        grid_row += 1
        self.log_text = ScrolledText(iframe, width=78, height=20 )
        self.log_text.grid(row=grid_row, column=0)

        self.guide_message = Tk.StringVar()
        self.guide_message_label = Tk.Label(lower_frame, bg='white', textvariable=self.guide_message)
        self.guide_message_label.pack(fill=Tk.X, padx=10, pady=10)
        self.update_guide_message()

        if False:
            fig = dplt.figure()
            ax = fig.gca()
            ax.plot( self.q, self.a, label='data' )
            ax.plot( self.qc, self.ac, label='fit data' )
            ax.legend()
            dplt.show()

    def create_figure(self, cframe):
        self.fig = fig = plt.figure( figsize=(7, 6) )
        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.ax = fig.gca()

        if self.q is None:
            self.add_dnd_bind()
            self.ax.text(0.5, 0.5, "Drag and drop\nor enter\nan input file.", alpha=0.3,
                        fontsize=36, ha='center', va='center')
        else:
            self.draw_curve()

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()

        self.ok_btn = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        self.ok_btn.pack(side=Tk.LEFT, padx=5, pady=5)
        self.ok_btn.pack_forget()

        self.cancel_btn = Tk.Button(box, text="Cancel", width=10, command=self.ask_cancel)
        self.cancel_btn.pack(side=Tk.LEFT, padx=5, pady=5)

        self.run_btn = Tk.Button(box, text="Run here", width=20, command=self.run_denss_thread)
        self.run_btn.pack(side=Tk.LEFT, padx=5, pady=5)

        self.submit_btn = Tk.Button(box, text="Run in background", width=20, command=self.submit_to_denss_manager)
        self.submit_btn.pack(side=Tk.LEFT, padx=5, pady=5)

        self.show_btn = Tk.Button(box, text="Show Manager", width=14, command=self.show_denss_manager)
        self.show_btn.pack(side=Tk.LEFT, padx=5, pady=5)

        self.edv_btn = Tk.Button(box, text="Electron Density Viewer", width=20, command=self.show_ed_viewer)
        self.edv_btn.pack(side=Tk.LEFT, padx=5, pady=5)

        self.update_button_states()

        # self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def is_ready(self):
        if self.is_memory_input:
            return not is_empty_val(self.out_folder.get())
        else:
            if is_empty_val(self.in_file.get()) or is_empty_val(self.out_folder.get()):
                return False
            else:
                return True

    def update_button_states(self):
        state = Tk.NORMAL if self.is_ready() else Tk.DISABLED
        for b in [self.illust_btn, self.fit_data_btn, self.run_btn, self.submit_btn]:
            b.config(state=state)

    def update_guide_message(self, message=None):

        if message is None:
            entry_names = []
            if self.is_memory_input:
                pass
            else:
                if is_empty_val(self.in_file.get()):
                    entry_names.append('"Input File"')

            if is_empty_val(self.out_folder.get()):
                entry_names.append('"Output Folder"')

            if len(entry_names) > 0:
                message = "You need to specify %s." % (" and ".join(entry_names))
            else:
                message = "You are ready to Run."

        self.guide_message.set(message)

    def ok(self, event=None):

        if not self.validate():
            self.initial_focus.focus_set() # put focus back
            return

        self.withdraw()
        self.update_idletasks()

        try:
            self.apply()
        finally:
            self.cancel()

        os.chdir(self.cwd_init)

    def ask_cancel(self):
        if self.thread is None:
            self.cancel()
            os.chdir(self.cwd_init)
            return

        yn = MessageBox.askyesno(
                'Cancel Confirmation',
                'Do you really want to stop the execution?',
                parent=self,
                )
        if not yn:
            return

        self.thread.terminate()
        self.thread.join()
        self.cancel_btn.pack_forget()
        self.ok_btn.pack(side=Tk.LEFT, padx=5, pady=5)
        os.chdir(self.cwd_init)
        logging.info('Execution canceled by user.')

    def prepare_from_args(self, q, a, e, infile_name):
        self.data = np.array([q, a, e]).T
        self.apply_fit_data(q, a, e, infile_name, use_memory_data=True)
        self.dmax.set(round(self.sasrec.D,2))

    def apply_fit_data(self, q, a, e, infile_name, use_memory_data=False, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.DENSS.DenssUtils as DenssUtils
            reload(DenssUtils)
        from .DenssUtils import fit_data_impl
        sasrec, work_info = fit_data_impl(q, a, e, file=infile_name, gui=True, use_memory_data=use_memory_data)
        self.sasrec = sasrec
        self.work_info = work_info
        self.qc = sasrec.qc
        self.ac = sasrec.Ic
        self.ec = sasrec.Icerr
        _, self.infile_name = os.path.split(infile_name)

    def on_file_entry(self):
        self.on_input(self.in_file.get())

    def on_folder_entry(self):
        self.update_button_states()
        self.update_guide_message()

    def on_fig_dnd(self, event):
        files = event.data.split(' ')
        file = files[0]
        self.in_file.set(file)
        self.on_input(file)

    def on_input(self, file):
        data, _ = np_loadtxt(file)
        q, a, e = data[:,0], data[:,1], data[:,2]
        self.in_file.set(file)
        self.q = q
        self.a = a
        self.e = e
        self.data = data
        self.apply_fit_data(q, a, e, file)
        self.dmax.set(round(self.sasrec.D,2))
        self.draw_curve()
        self.update_button_states()
        self.update_guide_message()

    def draw_curve(self):
        ax = self.ax
        ax.cla()
        ax.semilogy()
        ax.set_title( "Extrapolated profile: " + self.infile_name )
        ax.set_xlabel('Q')
        ax.set_ylabel('Intensity ($Log_{10}$)')
        ax.plot(self.q, self.a, label='input curve (LRF result)', color=get_color(1))
        ax.plot(self.qc, self.ac, label='modified curve with I(q=0)', color=get_color(2))
        ax.legend()
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def run_denss_thread(self):
        self.update_guide_message(message="Running DENSS.")
        self.run_btn.pack_forget()
        self.submit_btn.pack_forget()
        self.show_btn.pack_forget()
        self.edv_btn.pack_forget()
        self.fit_data_btn.config(state=Tk.DISABLED)     # avoid pressing this button while running denss
        out_folder = self.out_folder.get()
        mkdirs_with_retry(out_folder)
        self.root_logger  = logging.getLogger()
        log_file = out_folder + '/denss.log'
        self.root_logger.addHandler( ProgressLoghandler(self.log_text, log_file) )
        self.queue = queue.Queue()
        self.thread = Thread(
                        target=self.denss_thread,
                        name='DenssThread',
                        args=[]
                        )
        self.thread.start()
        self.counter = 0
        self.maximum = None
        self.last_index = self.progress_text.index(Tk.INSERT)
        self.after(200, self.update_progress_text)      # 200: need some time to have self.log_text written enough

    def update_progress_text(self):
        if self.maximum is None:
            text = self.log_text.get('1.0', Tk.END)
            """
            with open('debug.txt', 'w') as fh:
                fh.write(text)
            """
            m = log_maxnum_re.search(text)
            if m:
                self.maximum = int(m.group(1))
            else:
                self.maximum = 9999
            # print('maximum=', self.maximum)
            self.progress_bar['maximum'] = self.maximum

        try:
            while True:
                ret = self.queue.get(block=False)
                if ret[0] == 0:
                    if self.logger is not None:
                        self.logger.info("ret=%s", str(ret))

                    self.thread.join()
                    self.progress_bar['value'] = self.maximum
                    self.cancel_btn.pack_forget()
                    self.ok_btn.pack(side=Tk.LEFT, padx=5, pady=5)
                    self.edv_btn.pack(side=Tk.LEFT, padx=5, pady=5)
                    self.fit_data_btn.config(state=Tk.NORMAL)
                    self.update_guide_message(message="DENSS Done.")
                    return
                else:
                    message = ret[1]
                    if self.counter > 0:
                        self.progress_text.delete( self.last_index, Tk.INSERT )

                    self.last_index = self.progress_text.index(Tk.INSERT)
                    if len(message) > 0 and message[0] == '\r':
                        self.counter += 1
                        m = log_step_re.search(message[1:])
                        if m:
                            self.progress_bar['value'] = int(m.group(1))
                            self.progress_bar.update()
                    else:
                        # self.counter = 0
                        pass
                    self.progress_text.insert(Tk.INSERT, message)

        except queue.Empty:
            pass
        self.after(100, self.update_progress_text)

    def denss_thread(self, debug=True):
        os.chdir(self.out_folder.get())
        redirector = StdoutRedirector(self.queue)

        from molass_legacy.Env.EnvInfo import get_global_env_info
        if debug:
            from importlib import reload
            import molass_legacy.DENSS.DenssUtils as DenssUtils
            reload(DenssUtils)
        from .DenssUtils import run_denss_impl

        env_info = get_global_env_info(gpu_info=True)

        use_gpu = (env_info is not None and env_info.nvidiagpu_is_available)
        if use_gpu:
            logging.info('usign GPUs in DENSS.')

        dmax = self.dmax.get()
        # gui = False ok because redirector sets sys.stdout
        run_denss_impl(self.qc, self.ac, self.ec, dmax, self.infile_name, use_gpu=use_gpu)

        logging.info('DENSS thread finished.')
        sleep(0.3)      # to avoid hanging in pythonw.exe (while not sure if this is really effective)
        self.queue.put([0])

        os.chdir(self.cwd_init)

    def _cancel(self):
        os.chdir(self.cwd_init)
        self.cancel()

    def denss_fit_data(self, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.DENSS.DenssFitData as DenssFitData
            reload(DenssFitData)
        from .DenssFitData import DenssFitDataDialog
        dialog = DenssFitDataDialog(self, self.sasrec, self.work_info, self.infile_name, self.out_folder.get())
        dialog.show()
        if dialog.applied:
            sasrec = dialog.get_sasrec()
            self.dmax.set(round(sasrec.D,2))

    def submit_to_denss_manager(self):
        from .DenssManager import JobInfo
        from .DenssManagerDialog import show_manager_dialog
        dmax = self.dmax.get()
        job = JobInfo('denss', q=self.qc, a=self.ac, e=self.ec, dmax=dmax, infile_name=self.infile_name)
        def show_dialog():
            self._cancel()
            show_manager_dialog(self.parent, [job])
        self.parent.after(0, show_dialog)
        self.config(cursor='wait')
        self.denss_action = True

    def show_denss_manager(self):
        from .DenssManagerDialog import show_manager_dialog
        def show_dialog():
            self._cancel()
            show_manager_dialog(self.parent)
        self.parent.after(0, show_dialog)
        self.config(cursor='wait')
        self.denss_action = True

    def show_ed_viewer(self):
        print('show_ed_viewer')
        from molass_legacy.Saxs.EdViewer import EdViewer
        from molass_legacy.KekLib.OurMatplotlib import reset_to_default_style
        viewer = EdViewer(self)
        viewer.show()
        reset_to_default_style()

    def illustrate_dmax(self, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.DENSS.DmaxEstimation as DmaxEstimation
            reload(DmaxEstimation)
        from .DmaxEstimation import illustrate_dmax
        from molass_legacy.KekLib.DebugPlot import DialogWrapper

        with DialogWrapper(parent=self, window_title="Illustration"):
            fig, ax = dplt.subplots()
            ax.set_title("Dmax Estimation illustrated", fontsize=20)
            illustrate_dmax(ax, self.data)
            fig.tight_layout()



"""
    SecTheory.MolassProof.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import basinhopping
from datetime import datetime
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from .SecCF import simple_phi, shifted_phi, moving_zone_phi, two_site_phi
from .SecPDF import FftInvPdf, c, FftInvImpl

def proof_1():

    t = np.arange(300)
    npi = 100
    tpi = 1
    simple_pdf = FftInvPdf(simple_phi)

    t0 = 20
    shifted_pdf = FftInvPdf(shifted_phi)

    moving_pdf = FftInvPdf(moving_zone_phi)
    # N0 = (t0/3)**2
    N0 = 5000

    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))

        ax1.plot(t, c(t, npi, tpi), label="formula")
        ax1.plot(t, simple_pdf(t, npi, tpi), label="numerical")

        finv = FftInvImpl()
        w = finv.get_w()
        z = simple_phi(w, npi, tpi)
        u = finv.compute(t, z)
        ax1.plot(t, u, ":", label="inverse FFT")

        if False:
            w = np.fft.fftfreq(len(t))
            z = simple_phi(w, npi, tpi)
            u = np.fft.ifft(z)
            ax1.plot(t, u, ":", label="inverse np.fft")

        ax1.legend()

        ax2.plot(t, c(t - t0, npi, tpi), label="shifted formula")
        ax2.plot(t, shifted_pdf(t, npi, tpi, t0), label="shifted numerical")
        ax2.legend()

        ax3.plot(t, c(t - t0, npi, tpi), label="shifted formula")
        ax3.plot(t, moving_pdf(t, npi, tpi, N0, t0), label="modified numerical")
        ax3.legend()

        fig.tight_layout()
        plt.show()

def demo():
    from .SecCF import simple_phi, moving_zone_phi
    from .SecPDF import c, FftInvPdf

    t = np.arange(300)

    npi = 10
    tpi = 10
    simple_pdf = FftInvPdf(simple_phi)

    moving_pdf = FftInvPdf(moving_zone_phi)
    N0 = 6
    t0 = 6

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("Proof of Numerical Inversion", fontsize=20)
        ax1.set_title("Simple", fontsize=16)
        ax1.plot(t, c(t, npi, tpi), label="formula")
        ax1.plot(t, simple_pdf(t, npi, tpi), label="numerical")
        ax1.legend()
        ax2.set_title("Moving Zone considered", fontsize=16)
        ax2.plot(t, moving_pdf(t, npi, tpi, N0, t0), label="numerical")
        ax2.legend()
        fig.tight_layout()
        plt.show()



# coding: utf-8
"""
    FoleyGmm.py.

    adapted from:
        https://towardsdatascience.com/gaussian-mixture-modelling-gmm-833c88587c7f
        https://www.kaggle.com/dfoly1/gaussian-mixture-model

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import multivariate_normal as mvn
from sklearn.cluster import KMeans

class GMM:
    """ Gaussian Mixture Model

    Parameters
    -----------
        k: int , number of gaussian distributions
        seed: int, will be randomly set if None
        max_iter: int, number of iterations to run algorithm, default: 200

    Attributes
    -----------
       centroids: array, k, number_features
       cluster_labels: label for each data point

    """
    def __init__(self, C, n_runs):
        self.C = C # number of Guassians/clusters
        self.n_runs = n_runs

    def get_params(self):
        return (self.mu, self.pi, self.sigma)

    def calculate_mean_covariance(self, X, prediction):
        """Calculate means and covariance of different
            clusters from k-means prediction

        Parameters:
        ------------
        prediction: cluster labels from k-means
        X: N*d numpy array data points 

        Returns:
        -------------
        intial_means: for E-step of EM algorithm
        intial_cov: for E-step of EM algorithm

        """
        d = X.shape[1]
        labels = np.unique(prediction)
        self.initial_means = np.zeros((self.C, d))
        self.initial_cov = np.zeros((self.C, d, d))
        self.initial_pi = np.zeros(self.C)
        
        counter=0
        for label in labels:
            ids = np.where(prediction == label) # returns indices
            self.initial_pi[counter] = len(ids[0]) / X.shape[0]
            self.initial_means[counter,:] = np.mean(X[ids], axis = 0)
            de_meaned = X[ids] - self.initial_means[counter,:]
            Nk = X[ids].shape[0] # number of data points in current gaussian
            self.initial_cov[counter,:, :] = np.dot(self.initial_pi[counter] * de_meaned.T, de_meaned) / Nk
            counter+=1
        assert np.sum(self.initial_pi) == 1    
            
        return (self.initial_means, self.initial_cov, self.initial_pi)

    def _initialise_parameters(self, X):
        """Implement k-means to find starting
            parameter values.
            https://datascience.stackexchange.com/questions/11487/how-do-i-obtain-the-weight-and-variance-of-a-k-means-cluster

        Parameters:
        ------------
        X: numpy array of data points

        Returns:
        ----------
        tuple containing initial means and covariance
        _initial_means: numpy array: (C*d)
        _initial_cov: numpy array: (C,d*d)

        """
        n_clusters = self.C
        kmeans = KMeans(n_clusters= n_clusters, init="k-means++", max_iter=500, algorithm = 'auto')
        fitted = kmeans.fit(X)
        prediction = kmeans.predict(X)
        self._initial_means, self._initial_cov, self._initial_pi = self.calculate_mean_covariance(X, prediction)
        return (self._initial_means, self._initial_cov, self._initial_pi)

    def _e_step(self, X):
        """Performs E-step on GMM model

        Parameters:
        ------------
        X: (N x d), data points, m: no of features
        pi: (C), weights of mixture components
        mu: (C x d), mixture component means
        sigma: (C x d x d), mixture component covariance matrices

        Returns:
        ----------
        gamma: (N x C), probabilities of clusters for objects
        """
        N = X.shape[0] 
        self.gamma = np.zeros((N, self.C))

        for c in range(self.C):
            # Posterior Distribution using Bayes Rule
            self.gamma[:,c] = self.pi[c] * mvn.pdf(X, self.mu[c,:], self.sigma[c])

        # normalize across columns to make a valid probability
        gamma_norm = np.sum(self.gamma, axis=1)[:,np.newaxis]
        self.gamma /= gamma_norm

        return self.gamma

    def _m_step(self, X):
        """Performs M-step of the GMM
        We need to update our priors, our means
        and our covariance matrix.

        Parameters:
        -----------
        X: (N x d), data 
        gamma: (N x C), posterior distribution of lower bound 

        Returns:
        ---------
        pi: (C)
        mu: (C x d)
        sigma: (C x d x d)
        """
        N = X.shape[0] # number of objects
        C = self.gamma.shape[1] # number of clusters
        d = X.shape[1] # dimension of each object

        # responsibilities for each gaussian
        self.pi = np.mean(self.gamma, axis = 0)

        self.mu = np.dot(self.gamma.T, X) / np.sum(self.gamma, axis = 0)[:,np.newaxis]

        for c in range(C):
            x = X - self.mu[c, :] # (N x d)

            gamma_diag = np.diag(self.gamma[:,c])
            x_mu = np.matrix(x)
            gamma_diag = np.matrix(gamma_diag)

            sigma_c = x.T * gamma_diag * x
            self.sigma[c,:,:]=(sigma_c) / np.sum(self.gamma, axis = 0)[:,np.newaxis][c]

        return self.pi, self.mu, self.sigma

    def _compute_loss_function(self, X):
        """Computes lower bound loss function
        
        Parameters:
        -----------
        X: (N x d), data 
        
        Returns:
        ---------
        pi: (C)
        mu: (C x d)
        sigma: (C x d x d)
        """
        N = X.shape[0]
        C = self.gamma.shape[1]
        self.loss = np.zeros((N, C))

        for c in range(C):
            dist = mvn(self.mu[c], self.sigma[c],allow_singular=True)
            self.loss[:,c] = self.gamma[:,c] * (np.log(self.pi[c]+0.00001)+dist.logpdf(X)-np.log(self.gamma[:,c]+0.000001))
        self.loss = np.sum(self.loss)
        return self.loss

    def fit(self, X):
        """Compute the E-step and M-step and
            Calculates the lowerbound
        
        Parameters:
        -----------
        X: (N x d), data 
        
        Returns:
        ----------
        instance of GMM
        
        """
        
        d = X.shape[1]
        self.mu, self.sigma, self.pi =  self._initialise_parameters(X)

        try:
            for run in range(self.n_runs):  
                self.gamma  = self._e_step(X)
                self.pi, self.mu, self.sigma = self._m_step(X)
                loss = self._compute_loss_function(X)

                if run % 10 == 0:
                    print("Iteration: %d Loss: %0.6f" %(run, loss))

        except Exception as e:
            print(e)

        return self

    def predict(self, X):
        """Returns predicted labels using Bayes Rule to
        Calculate the posterior distribution
        
        Parameters:
        -------------
        X: ?*d numpy array
        
        Returns:
        ----------
        labels: predicted cluster based on 
        highest responsibility gamma.
        
        """
        labels = np.zeros((X.shape[0], self.C))
        
        for c in range(self.C):
            labels [:,c] = self.pi[c] * mvn.pdf(X, self.mu[c,:], self.sigma[c])
        labels  = labels .argmax(1)
        return labels 

    def predict_proba(self, X):
        """Returns predicted labels
        
        Parameters:
        -------------
        X: N*d numpy array
        
        Returns:
        ----------
        labels: predicted cluster based on 
        highest responsibility gamma.
        
        """
        post_proba = np.zeros((X.shape[0], self.C))

        for c in range(self.C):
            # Posterior Distribution using Bayes Rule, try and vectorise
            post_proba[:,c] = self.pi[c] * mvn.pdf(X, self.mu[c,:], self.sigma[c])

        return post_proba



"""
    Theory.WLS.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def npWLS(x, y, w=None):
    pass

def npWLS_spike():
    x = np.array([0, 1, 2, 3])
    y = np.array([-1, 0.2, 0.9, 2.1])
    A = np.vstack([x, np.ones(len(x))]).T
    m, c = np.linalg.lstsq(A, y, rcond=None)[0]

    fig, ax = plt.subplots()
    ax.plot(x, y, 'o', label='Original data', markersize=10)
    ax.plot(x, m*x + c, 'r', label='Fitted line')
    ax.legend()
    plt.show()



# coding: utf-8
"""
    HplcSimulator.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as mplt
from matplotlib.gridspec import GridSpec
import mpl_toolkits.mplot3d.axes3d as p3
from mpl_toolkits.mplot3d import proj3d
import matplotlib.animation as animation
from shapely.geometry import LineString, Point
from descartes import PolygonPatch
import molass_legacy.KekLib.DebugPlot as plt

"""
From:
Add cylinder to plot
https://stackoverflow.com/questions/26989131/add-cylinder-to-plot
"""

RADIUS = 0.3
MOVE_DELTA = 0.1
TITLE_FONTSIZE = 32

COLUMN_CENTER_Y = 0.5
COLUMN_CENTER_Z = 0.5
COLUMN_LENGTH = 3
COUNTER_LENGTH = 0.1
START_LENGTH = 0.1

def data_for_cylinder_along_z(center_x, center_y, radius, min_height, max_height, res_h=50, res_c=50):
    z = np.linspace(min_height, max_height, res_h)
    theta = np.linspace(0, 2*np.pi, res_c)
    theta_grid, z_grid=np.meshgrid(theta, z)
    x_grid = radius*np.cos(theta_grid) + center_x
    y_grid = radius*np.sin(theta_grid) + center_y
    return x_grid,y_grid,z_grid

"""
Generate a random point within a circle (uniformly)
https://stackoverflow.com/questions/5837572/generate-a-random-point-within-a-circle-uniformly
"""
def uniform_random_disc(n, radius):
    r = radius * np.sqrt(np.random.rand(n))
    theta = np.random.rand(n) * 2 * np.pi
    return r * np.cos(theta), r * np.sin(theta)

class Cylinder:
    def __init__(self, xmin, xmax, radius):
        self.xmin = xmin
        self.xmax = xmax
        self.radius = radius
        self.ycenter = COLUMN_CENTER_Y
        self.zcenter = COLUMN_CENTER_Z

class PaticleGroup:
    def __init__(self, n, cylinder, color=None, speed=0):
        self.n = n
        self.r = cylinder.radius
        self.cy = cylinder.ycenter
        self.cz = cylinder.zcenter
        self.xmin = cylinder.xmin
        self.xmax = cylinder.xmax
        self.color = color
        self.speed = speed
        th = -np.pi/2
        c = np.cos(th)
        s = np.sin(th)
        self.rotation = np.array([[c, -s], [s, c] ])
        self.set_init_positions()

    def set_init_positions(self):
        r = self.r
        cy = self.cy
        cz = self.cz
        self.circle = Point(cy, cz).buffer(r)
        n = self.n
        self.x = self.xmin + (self.xmax - self.xmin) * np.random.rand(n)
        y_, z_ = uniform_random_disc(n, r)
        self.y = cy + y_
        self.z = cz + z_

    def plot(self, ax):
        self.artists, = ax.plot(self.x, self.y, self.z, 'o', color=self.color, markersize=1, alpha=0.5)

    def get_artists(self):
        return [self.artists]

    def reset(self):
        self.set_init_positions()
        self.artists.set_data(self.x, self.y)
        self.artists.set_3d_properties(self.z)

    def compute_bounced_coord(self, ty, tz, ey, ez):
        p = np.array([(ey - ty), (ez - tz)]).T
        v = np.array([(self.cy - ty), (self.cz - tz)]).T/self.r
        u = np.dot(self.rotation, v)
        W = np.array([u, v]).T
        W_ = np.linalg.inv(W)
        q_ = np.dot(W_, p)
        q = np.dot(W, np.array([q_[0], -q_[1]]))
        return ty+q[0], tz+q[1]

    def random_move(self, delta):
        n = self.n
        self.x += (np.random.rand(n) - 0.5) * delta + self.speed * delta
        y = self.y + (np.random.rand(n) - 0.5) * delta
        z = self.z + (np.random.rand(n) - 0.5) * delta

        r = self.r
        cy = self.cy
        cz = self.cz
        circle = self.circle
        for k in np.where((y - cy)**2 + (z - cz)**2 > r**2)[0]:
            segment = LineString([(self.y[k], self.z[k]), (y[k], z[k])])
            coords = circle.intersection(segment).coords
            if len(coords) == 0:
                # this case does occur
                segment = LineString([(cy, cz), (y[k], z[k])])
                coords = circle.intersection(segment).coords
                if len(coords) == 0:
                    continue

            q = self.compute_bounced_coord(*coords[-1], y[k], z[k])

            if False:
                plt.push()
                fig = plt.figure()
                ax = fig.gca()
                ax.set_xlim(0, 1)
                ax.set_ylim(0, 1)
                ax.set_aspect('equal')
                patch = PolygonPatch(circle, alpha=0.1)
                ax.add_patch(patch)
                for c in coords:
                    ax.plot(*c, 'o')
                ax.plot(y[k], z[k], 'o', color='red')
                ax.plot([cy, c[0]], [cz, c[1]], color='yellow')
                ax.plot(*q, 'o', color='green')
                plt.show()
                plt.pop()
                exit()

            y[k] = q[0]
            z[k] = q[1]

        self.y = y
        self.z = z
        self.artists.set_data(self.x, self.y)
        self.artists.set_3d_properties(self.z)

    def get_num_detected(self):
        return np.sum(np.logical_and(self.x > 2.9, self.x < 3.0))

class HplcSimulator:
    def __init__(self, num_particles=5000, num_frames=250, start_length=START_LENGTH, plt=mplt):
        self.num_particles = num_particles
        self.num_frames = num_frames
        self.start_length = start_length
        self.plt = plt
        self.fig = fig = plt.figure(figsize=(18,10))
        gs = GridSpec(2, 3)
        ax0 = fig.add_subplot(gs[0,:], projection='3d')
        # print('dir(ax1)=', dir(ax1))
        ax1 = fig.add_subplot(gs[1,0])
        ax2 = fig.add_subplot(gs[1,1])
        ax3 = fig.add_subplot(gs[1,2])
        self.axes = ax0, ax1, ax2, ax3

    def play(self, save_only=False):
        ax0, ax1, ax2, ax3 = self.axes

        Xc, Yc, Zc = data_for_cylinder_along_z(COLUMN_CENTER_Y, COLUMN_CENTER_Z, RADIUS, 0, COLUMN_LENGTH)
        entire_column = ax0.plot_surface(Zc, Xc, Yc, alpha=0.1)
        Xc, Yc, Zc = data_for_cylinder_along_z(COLUMN_CENTER_Y, COLUMN_CENTER_Z, RADIUS, COLUMN_LENGTH-COUNTER_LENGTH, COLUMN_LENGTH, res_h=5)
        counter_column = ax0.plot_surface(Zc, Xc, Yc, alpha=0.5, color='yellow')

        N = self.num_particles
        ax0.set_title("Conceptully Simulated HPLC Column (%d red, %d blue Particles)" % (N, N), fontsize=TITLE_FONTSIZE, y=1.1)
        ax1.set_title("Number of Particles", fontsize=TITLE_FONTSIZE)
        ax2.set_title("UV Absorbance", fontsize=TITLE_FONTSIZE)
        ax3.set_title("Xray Scattering", fontsize=TITLE_FONTSIZE)

        ax0.set_xlim3d(0.6, 2.5)
        ax0.set_ylim3d(0, 1)
        shrink = 0.25
        ax0.set_zlim3d(shrink, 1-shrink)
        ax0.set_axis_off()
        ax0.view_init(0, -90)

        tx, ty, _ = proj3d.proj_transform(2.95, 0.5, 0.2, ax0.get_proj())
        ax0.annotate("Counter Ring", xy=(tx, ty), xytext=(tx, ty-0.02), ha='center', arrowprops=dict( arrowstyle="->", color='k' ) )

        start_column = Cylinder(0, self.start_length, RADIUS)
        red_particles = PaticleGroup(self.num_particles, start_column, color='red', speed=0.2)
        red_particles.plot(ax0)
        blue_particles = PaticleGroup(self.num_particles, start_column, color='blue', speed=0.4)
        blue_particles.plot(ax0)

        hx = np.arange(self.num_frames)
        zeros = np.zeros(self.num_frames)
        hy1 = zeros.copy()
        hy2 = zeros.copy()
        hy = zeros.copy()

        hist1, = ax1.plot(hx, hy1, ':', color='red', label='red particles')
        hist2, = ax1.plot(hx, hy2, ':', color='blue', label='blue particles')
        hist, = ax1.plot(hx, hy, color='green', alpha=0.2, label='total')
        ax1.legend()

        artists = []
        hg_artists = [hist1, hist2, hist]
        hg_y_s = [hy1, hy2, hy]
        artists += hg_artists

        ax1.set_xlim(0, self.num_frames)
        ax1.set_ylim(0, 0.2*self.num_particles)

        def scaled_plot(ax, scale1, scale2):
            y1 = hy1*uv_scale1
            y2 = hy2*uv_scale2
            y = y1 + y2
            ret1, = ax.plot(hx, y1, ':', color='red', label='red particles')
            ret2, = ax.plot(hx, y2, ':', color='blue', label='blue particles')
            ret, = ax.plot(hx, y, color='green', alpha=0.2, label='total')
            ax.legend()
            return [ret1, ret2, ret], [y1, y2, y]

        uv_scale1 = 0.1/self.num_frames
        uv_scale2 = 0.2/self.num_frames

        ax2.set_xlim(0, self.num_frames)
        ax2.set_ylim(0, 0.2*self.num_particles*max(uv_scale1, uv_scale2))

        uv_artists, uv_y_s = scaled_plot(ax2, uv_scale1, uv_scale2)
        artists += uv_artists

        xr_scale1 = 0.08/self.num_frames
        xr_scale2 = 0.05/self.num_frames

        ax3.set_xlim(0, self.num_frames)
        ax3.set_ylim(0, 0.2*self.num_particles*max(xr_scale1, xr_scale2))

        xr_artists, xr_y_s = scaled_plot(ax3, xr_scale1, xr_scale2)
        artists += xr_artists

        self.fig.tight_layout()
        self.fig.subplots_adjust(top=0.92, hspace=0.2)

        def reset():
            red_particles.reset()
            blue_particles.reset()
            for k, a in enumerate(hg_artists):
                hg_y_s[k] = y = zeros.copy()
                a.set_data(hx, y)
            for k, a in enumerate(uv_artists):
                uv_y_s[k] = y = zeros.copy()
                a.set_data(hx, y)
            for k, a in enumerate(xr_artists):
                xr_y_s[k] = y = zeros.copy()
                a.set_data(hx, y)
            return red_particles.get_artists() + blue_particles.get_artists() + artists

        def animate(i):
            red_particles.random_move(MOVE_DELTA)
            blue_particles.random_move(MOVE_DELTA)
            c1 = red_particles.get_num_detected()
            c2 = blue_particles.get_num_detected()

            def set_data_to_artists(artists_, y_s, scale1, scale2):
                y1, y2, y = y_s
                y1[i] = c1*scale1
                y2[i] = c2*scale2
                y[i] = y1[i] + y2[i]
                for a, ay in zip(artists_, y_s):
                    a.set_data(hx, ay)

            set_data_to_artists(hg_artists, hg_y_s, 1, 1)
            set_data_to_artists(uv_artists, uv_y_s, uv_scale1, uv_scale2)
            set_data_to_artists(xr_artists, xr_y_s, xr_scale1, xr_scale2)

            return red_particles.get_artists() + blue_particles.get_artists() + artists

        self.anim = animation.FuncAnimation(self.fig, animate, frames=self.num_frames, blit=True, init_func=reset)

        if save_only:
            pass
        else:
            self.plt.show()

    def save(self, file, **kwargs):
        self.anim.save(file, **kwargs)




"""
    KekLib/IntegrateUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.integrate import quad, quad_vec

"""
copied from https://stackoverflow.com/questions/5965583/use-scipy-integrate-quad-to-integrate-complex-numbers
"""
def complex_quadrature(func, a, b, **kwargs):
    def real_func(x):
        return np.real(func(x))
    def imag_func(x):
        return np.imag(func(x))
    real_integral = quad(real_func, a, b, **kwargs)
    imag_integral = quad(imag_func, a, b, **kwargs)
    return (real_integral[0] + 1j*imag_integral[0], real_integral[1:], imag_integral[1:])

"""
quad_vec version
"""
def complex_quadrature_vec(func, a, b, **kwargs):
    def real_func(x):
        return np.real(func(x))
    def imag_func(x):
        return np.imag(func(x))
    real_integral = quad_vec(real_func, a, b, **kwargs)
    imag_integral = quad_vec(imag_func, a, b, **kwargs)
    return (real_integral[0] + 1j*imag_integral[0], real_integral[1:], imag_integral[1:])


# coding: utf-8
"""
    ScipyUtils.py

    Copyright (c) 2018-2021, Masatsuyo Takahashi, KEK-PF
"""

import numpy as np
from scipy.special import erfcx

"""
    inverse of erfcx (scaled complementary error function)
    https://jp.mathworks.com/matlabcentral/answers/302915-inverse-of-erfcx-scaled-complementary-error-function#
"""
def erfcxinv_impl(y):
    # erfcx inverse, for y no larger than 1.
    # for y <= 1, use the large x approximation for a starting value.
    k = y <= 1
    x = np.zeros(len(y));
    x[k] = 1/(y[k]*np.sqrt(np.pi))
    # for y > 1, use exp(x^2) as a very rough approximation
    # to erfcx
    _k = y > 1
    x[_k] = -np.sqrt(np.log(y[_k]));
    for n in range(7):
        x = x - (erfcx(x) - y)/(2*x*erfcx(x) - 2/np.sqrt(np.pi))
    return x

def erfcxinv(y):
    if np.isscalar(y):
        return erfcxinv_impl(np.array([y]))[0]
    else:
        return erfcxinv_impl(y)

def get_spline(x, y, num_knots=8):
    from scipy.interpolate import LSQUnivariateSpline
    knots = np.linspace(x[0], x[-1], num_knots + 2)
    return  LSQUnivariateSpline(x, y, knots[1:-1], ext=3)



"""
    ModelParams.ParamsSheetBase.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk

"""
    task: let this class have more in common
"""
class ParamsSheetBase(Tk.Frame):
    def __init__(self, parent, params, dsets, optimizer):
        self.parent = parent
        self.active_indeces = None
        self.num_params = len(params)
        Tk.Frame.__init__(self, parent)

    def enable_selection(self):
        self.sheet.enable_bindings(bindings="toggle_select")

    def enable_copy(self):
        self.sheet.enable_bindings()

    def set_selection(self, sel_list):
        for k in sel_list:
            r, c = self.get_cell_address(k)
            self.sheet.toggle_select_cell(r, c)

    def set_params_addr(self, k, addr):
        self.params_addr[k] = addr
        self.params_addr_inv[addr] = k

    def get_cell_address(self, k):
        return self.params_addr[k]

    def get_selection(self):
        ret_list = []
        for c in self.sheet.get_selected_cells():
            k = self.params_addr_inv.get(c)
            if k is None:
                import molass_legacy.KekLib.CustomMessageBox as MessageBox
                MessageBox.showerror("Invalid Cell", "Invalid Cell", parent=self.parent)
                raise ValueError
            else:
                ret_list.append(k)

        if len(ret_list) != 2:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            MessageBox.showerror("Number of Cells", "You must select exactly two cells.", parent=self.parent)
            raise ValueError

        return sorted(ret_list)

    def save_as(self, file):
        with open(file, "w") as fh:
            for k, row in enumerate(self.data_list, start=1):
                fh.write(",".join(row) + "\n")
                if k >= self.num_valid_rows:
                    break

    def get_active_indeces(self, state_info):
        # print("get_active_indeces: params_addr=", self.params_addr)
        if self.active_indeces is None:
            indeces = state_info.get_active_indeces()
            if len(indeces) == 0:
                indeces = list(range(self.num_params))
            self.active_indeces = indeces
        return self.active_indeces


"""
    Models/RateTheory/PureEDM_Demo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
def demo(in_folder, sd):
    from importlib import reload
    import Models.RateTheory.PureEDM
    reload(Models.RateTheory.PureEDM)
    from molass_legacy.Models.RateTheory.PureEDM import guess_multiple, edm_impl

    ecurve = sd.get_xray_curve()
    x = ecurve.x.copy()
    y = ecurve.y

    num_peaks = len(ecurve.peak_info) + 1

    if False:
        fh = open("moments-params.csv", "w")
    else:
        fh = None

    guess_multiple(x, y, num_peaks, fh=fh, debug=True)

    if fh is not None:
        fh.close()



"""
    SecTheory.DiporeSecSimulation.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
from matplotlib.gridspec import GridSpec

def demo(use_cf=True):
    import molass_legacy.KekLib.DebugPlot as plt
    if use_cf:
        from SecTheory.SecCF import sdm_dipore
        from SecTheory.SecPDF import FftInvPdf
        dipore_pdf = FftInvPdf(sdm_dipore)
    else:
        # 
        assert False

    x = np.arange(300)

    me = 0.1
    mp = 1.5
    rp = 76
    rg = 35
    rho = rg/rp
    N = 171
    T = 0.63
    t0 = 100
    sigma0 = 10
    N0 = (t0/sigma0)**2
    x0 = -20

    rg_array = np.array([60, 50, 40, 30, 20, 10, 5])

    y_list = []
    for rg_ in rg_array:
        rho_ = rg_/rp
        np_ = N*(1 - rho_)**me
        tp_ = T*(1 - rho_)**mp
        if use_cf:
            y = dipore_pdf(x, 0.5*np_, tp_, 0.5*np_, tp_, N0, t0, x0)
        else:
            assert False
        y_list.append(y)

    with plt.Dp():
        fig, ax = plt.subplots()

        for y, rg_ in zip(y_list, rg_array):
            ax.plot(x, y, label="$R_g=%.3g$" % rg_)

        ax.legend()
        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    demo()



"""
    UV/XrayProportional.py

    Copyright (c) 2018-2023, SAXS Team, KEK-PF
"""
import os
import numpy as np
from scipy.interpolate import UnivariateSpline
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Elution.CurveUtils  import get_xray_elution_vector
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy.KekLib.NumpyUtils import np_loadtxt
from ScatteringBaseCorrector import compute_baseline_using_LPM_impl
from molass_legacy.Mapping.SingleComponent import SingleComponent
import molass_legacy.KekLib.DebugPlot as plt

def make_xray_proportinal_uv_data_impl(xray_array, ivector=None):
    this_dir = os.path.dirname( os.path.abspath( __file__ ) )
    uv_ridge_data, _ = np_loadtxt( this_dir + '/UV.dat' )

    lvector_xp     = uv_ridge_data[:, 0]
    uv_ridge_curve  = uv_ridge_data[:, 1]

    if ivector is None:
        qvector = xray_array[0, :, 0]
        ivector, _ = get_xray_elution_vector( qvector, xray_array )

        if get_setting('apply_backsub') == 0:
            ivector_    = ivector
        else:
            ivector_    = copy.deepcopy(ivector)
            compute_baseline_using_LPM_impl( 0, 2, np.arange( len(ivector_) ), ivector_ )
    else:
        ivector_ = ivector

    x_curve = ElutionCurve( ivector_ )
    elution = x_curve.y / x_curve.max_y
    uv_array_xp    = np.dot( uv_ridge_curve.reshape( len(uv_ridge_curve), 1 ), elution.reshape( 1, len(elution) ) )
    uv_file_xp     = 'xray_proportinal.dat'

    return lvector_xp, uv_array_xp, uv_file_xp

def make_proportional_mapping_info_impl(mapper, opt_params, debug=False):
    mapper.opt_params = opt_params

    mapper.a_base_adjustment  = 0
    mapper.x_base_adjustment  = 0
    mapper.A_init = 1
    mapper.B_init = 0
    mapper.a_curve = mapper.make_a_curve(opt_params)
    mapper.x_curve = mapper.make_x_curve(opt_params)
    mapper.logger.info("made proportional curves with len(a_curve.x)=%d, len(x_curve.x)=%d", len(mapper.a_curve.x), len(mapper.x_curve.x))

    scale = np.max(mapper.a_vector) / np.max(mapper.x_vector)

    if len(mapper.x_x) == len(mapper.a_x):
        mapper.a_base = mapper.x_base * scale
    else:
        """
            previously, len(mapper.x_x) and len(mapper.a_x) were the same.
            the situation seems to have been changed probably due to Python (extension modules) updates.
            therefore, adjusting the difference using a spline
        """
        a_base_spline = UnivariateSpline(mapper.x_x, mapper.x_base * scale, s=0)
        mapper.a_base = a_base_spline(mapper.a_x)

    mapper.x_curve_y_adjusted = mapper.x_curve.y

    if debug:
        for v in [mapper.a_x, mapper.x_x, mapper.a_vector, mapper.a_base, mapper.x_base]:
            print(len(v))
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("make_proportional_mapping_info_impl debug")
            ax1.plot(mapper.a_vector)
            ax1.plot(mapper.a_base)

            ax2.plot(mapper.x_curve_y_adjusted)
            ax2.plot(mapper.x_curve.y)
            ax2.plot(mapper.x_base, color='red')
            fig.tight_layout()
            plt.show()

    mapper.determine_mapping_ranges()
    make_proportional_opt_results_impl(mapper)
    mapper.set_peak_eval_ranges()
    mapper.std_diff = 0
    mapper.opt_results

    mapper.mapped_vector = mapper.x_curve_y_adjusted
    mapper.mapped_spline  = UnivariateSpline( mapper.x_x, mapper.mapped_vector, s=0, ext=3 )
    mapper.inv_mapped_boundaries = mapper.x_curve.boundaries

    mapper.in_xray_adjustment_mode = False
    mapper.scomp = SingleComponent(mapper)
    mapper.logger.info("made xray-proportional mapping info with len(a_base)=%d, len(a_vector)=%d", len(mapper.a_base), len(mapper.a_vector))
    assert len(mapper.a_base) == len(mapper.a_vector)

def make_proportional_opt_results_impl(mapper):
    A = 1
    B = 0
    mapper.opt_results = []
    mapper.map_params = (A, B)
    S = 1   # TODO
    mapper.scale_params = [S] * len( mapper.mapping_ranges)
    for i, S in enumerate(mapper.scale_params):
        mapper.opt_results.append( [ A, B, S ] )



# coding: utf-8
"""
    QuantileRegression.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
# from QuantileRegressionKeras import Model
from QuantileRegressionSklearn import Model
import molass_legacy.KekLib.DebugPlot as plt

class QuantileRegression:
    def __init__(self, sd):
        array = sd.intensity_array

        N = 11

        q = array[0,:,0]
        qq = np.hstack([q]*N)

        i = 340
        curve_y = array[i,:,1]
        H = N//2
        yy = np.hstack([ array[k,:,1] for k in range(i-H, i+H+1) ])

        model = Model()
        model.fit(qq, yy)
        learned_y = model.predict(q)

        fig = plt.figure(figsize=(16,6))
        ax1 = fig.add_subplot(121)
        ax2 = fig.add_subplot(122)
        ax1.plot(sd.xray_curve.y, label='smp')

        for j in range(0, array.shape[1], 100):
            ax1.plot(array[:,j,1], label='j=%d' % j)

        ax2.plot(q, curve_y, label='i=%d' % i)
        ax2.plot(q, learned_y, label='learned')

        ax1.legend()
        ax2.legend()
        fig.tight_layout()
        plt.show()



"""

    ZeroExtrapolationResultBook.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF

"""
import os
import numpy                as np
import logging
# from openpyxl               import Workbook
from openpyxl.chart         import LineChart, ScatterChart, Reference, Series
from openpyxl.chart.series_factory  import SeriesFactory
from openpyxl.chart.error_bar       import ErrorBars
from openpyxl.chart.data_source     import NumDataSource, NumData, NumVal
from openpyxl.chart.layout  import Layout, ManualLayout
from OpenPyXlUtil           import save_allowing_user_reply, LINE_WIDTH
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
from HtmlColorNames         import *
from molass_legacy.Test.TesterLogger      import write_to_tester_log
from molass_legacy._MOLASS.SerialSettings         import get_setting
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

CHART_WIDTH_ZX_P    = 21
CHART_HEIGHT_ZX_P   = 14
CHART_WIDTH_ZX_I    = 62.5
CHART_HEIGHT_ZX_I   = 24
PARAM_A_NAME    = 'A(q) - Scattering Intensity without Interparticle Effects (1.0 mg/ml)'
PARAM_B_NAME    = 'B(q) - Interparticle Effects'
DUMP_INPUT      = False
ADD_ERROR_BARS  = True
ATSAS_COLOR     = darkgray

"""
    borrowed from https://github.com/uskysd/openpyxl-errorbar
"""
def list2errorbars(plus, minus, errDir='y', errValType='cust'):
    "Returns ErrorBar from lists of error values"

    #Convert to list of NumVal
    numvals_plus = [NumVal(i, None, v=x) for i,x in enumerate(plus)]
    numvals_minus = [NumVal(i, None, v=x) for i,x in enumerate(minus)]

    # Convert to NumData
    nd_plus = NumData(pt=numvals_plus)
    nd_minus = NumData(pt=numvals_minus)

    # Convert to NumDataSource
    nds_plus = NumDataSource(numLit=nd_plus)
    nds_minus = NumDataSource(numLit=nd_minus)

    return ErrorBars(plus=nds_plus, minus=nds_minus, errDir=errDir, errValType=errValType)

def set_layout_small( chart, adjust=0 ):
    chart.layout = Layout(
        ManualLayout(
        x=0.1+adjust, y=0.1,
        h=0.8,  w=0.76+adjust,
        xMode="edge",
        yMode="edge",
        )
    )

def set_layout_large( chart, adjust=0, r_adjust=0 ):
    chart.layout = Layout(
        ManualLayout(
        x=0.04-adjust, y=0.1,
        h=0.8,  w=0.92+adjust+r_adjust,
        xMode="edge",
        yMode="edge",
        )
    )

def create_extrapolation_params_chart_( ws, ud, num_rows, row_offset, col_offset, j, param_name, exstra_col, no_orig_param_array, r_adjust=0, overlay=False ):
    c_ = ScatterChart()
    c_.title = param_name
    c_.style = 13
    c_.y_axis.title = 'log( I )' if j == 0 else 'Intensity'
    c_.x_axis.title = 'Q(Å⁻¹)'

    min_row = row_offset + 1
    max_row = row_offset + num_rows + 1

    xvalues = Reference(ws, min_col=1,
                            min_row=min_row + 1,    # must not include column title
                            max_row=max_row)

    if no_orig_param_array:
        min_col = col_offset + j * 2
        max_col = min_col + exstra_col
    else:
        min_col = col_offset + j * 3
        max_col = min_col + 1 + exstra_col

    if overlay:
        colors = [ red, blue ]
    else:
        primary_color = red if ud == 0 else blue
        if no_orig_param_array:
            colors = [ primary_color ]
        else:
            # TODO: test this case
            colors = [ lightgray, primary_color ]
        if exstra_col == 1:
            # colors.append( darkseagreen )
            colors.append( ATSAS_COLOR )

    for col in range( min_col, max_col+1 ):
        values = Reference(ws, min_col=col, min_row=min_row, max_row=max_row)
        series = Series(values, xvalues, title_from_data=True)
        c_.series.append(series)

    if j == 0:
        c_.y_axis.scaling.logBase = 10

    c_.width    = CHART_WIDTH_ZX_I
    c_.height   = CHART_HEIGHT_ZX_I

    # Style the lines

    for i, color in enumerate(colors):
        s_ = c_.series[i]
        # s_.marker.symbol = "circle"
        s_.marker.graphicalProperties.solidFill         = color
        s_.marker.graphicalProperties.line.solidFill    = color
        s_.graphicalProperties.line.solidFill           = color
        width = LINE_WIDTH    # in EMUs
        s_.graphicalProperties.line.width               = width

    set_layout_large( c_, r_adjust=r_adjust )

    return c_

ERROR_SCALE_FOR_AXIS = 2

def get_max( a, b, error, atsas=False, max_limit_ratio=1.5 ):
    if atsas:
        error = 0
    else:
        error *= ERROR_SCALE_FOR_AXIS
    if a is None:
        if b is None:
            return None
        else:
            return b + error
    else:
        if b is None:
            return a
        else:
            b += error
            if atsas:
                if a < b and b < a*max_limit_ratio:
                    # do not adopt too large b
                    return b
                else:
                    return a
            else:
                if a < b:
                    return b
                else:
                    return a

def get_min( a, b, error, atsas=False ):
    if atsas:
        error = 0
    else:
        error *= ERROR_SCALE_FOR_AXIS
    if a is None:
        if b is None:
            return None
        else:
            return b - error
    else:
        if b is None:
            return a
        else:
            b -= error
            if a < b:
                return a
            else:
                return b

def divide( a, b, ret ):
    if a is None:
        return ret
    else:
        return a/b

class ZeroExtrapolationResultBook:
    def __init__( self, wb, ws,
                    mjn, indeces_, c_rg_iz, qvector, data_matrix, j0,
                    ze_array, param_array, orig_param_array=None,
                    guinier_boundary=None,
                    almerge_result=None, atsas_result=None,
                    boundary_j=None, parent=None, lrf_info=None ):
        m, j, num_peaks_to_exec, peak_num_ranges = mjn
        # print( 'ZeroExtrapolationResultBook: mjn=', mjn )
        self.logger = logging.getLogger( __name__  )
        self.parent = parent
        self.doing_sec = parent.doing_sec
        self.need_bq = lrf_info.need_bq()
        self.lrf_boundary_j = lrf_info.boundary_j

        if almerge_result is None:
            exz_array       = None
            overlap_from_max  = None
        else:
            exz_array       = almerge_result.exz_array
            overlap_from_max  = almerge_result.overlap_from_max

        indeces = j0 + np.array(indeces_)

        if DUMP_INPUT and m == 0 and j == 0:
            from molass_legacy.Reports.ReportUtils import dump_zx_report_args
            dump_zx_report_args( m, j, indeces, c_rg_iz, qvector, data_matrix, ze_array, param_array, orig_param_array, exz_array )

        self.m  = m
        self.ud = j
        self.num_peaks_to_exec = num_peaks_to_exec
        self.guinier_boundary = guinier_boundary
        self.indeces = indeces
        self.wb = wb
        self.ws = ws
        num_intensity_points = param_array.shape[0]
        self.num_rows = num_intensity_points + 1
        paren = '' if num_peaks_to_exec == 1 else '(%d)' % ( m + 1 )
        range_type = get_setting('range_type')
        if self.doing_sec:
            if peak_num_ranges == 1 and range_type < 5:
                comp = 'Both-side'
            else:
                comp = 'Asc-side' if self.ud == 0 else 'Desc-side'
        else:
            comp = 'Nat-comp' if self.m == 0 else 'Unf-comp'

        ws.title = '%s Extrapolation%s' % ( comp, paren )

        ws.append( [ 'Elution №', 'Input', '_MOLASS',
                        'Input', '_MOLASS', 'ATSAS',
                        'Input', '_MOLASS', 'ATSAS',
                        'Rg-error', 'x-Rg-error', 'atsas-x-Rg-error',
                        'I(0)/Conc-error', 'x-I(0)/Conc-error', 'atsas-x-I(0)/Conc-error' ] )
        self.nx_rows = nx_rows = len( c_rg_iz[0][0] )
        self.zx_rows = zx_rows = len( c_rg_iz[0][1] )
        print( 'nx_rows=', nx_rows, ', zx_rows=', zx_rows )
        assert( nx_rows <= zx_rows )

        self.orig_param_array = orig_param_array
        self.exz_array      = exz_array
        self.overlap_from_max = overlap_from_max
        self.boundary_j     = boundary_j

        c_vector    = c_rg_iz[0][0]
        self.c_max  = np.max( c_vector )
        zx_only_rows = zx_rows - nx_rows

        # TODO: c_rg_iz refactoring

        rg_min = None
        rg_max = None
        iz_min = None
        iz_max = None

        rg_errors   = []
        rgx_errors  = []
        atsas_rgx_errors  = []
        iz_errors   = []
        izx_errors  = []
        atsas_izx_errors = []
        if self.ud == 0:
            for i in range(zx_rows):
                if atsas_result is None or i > 0:
                    atsas_Rg        = None
                    atsas_Iz        = None
                else:
                    atsas_Rg        = atsas_result.Rg
                    # atsas_Rg_error  = atsas_result.Rg_stdev
                    # atsas_Iz should have been scaled by the max concentration
                    atsas_Iz        = divide( atsas_result.I0, self.c_max, None )
                    # atsas_Iz_error  = divide( atsas_result.I0_stdev, self.c_max, 0 )

                atsas_Rg_error  = 0   # or np.nan?
                atsas_Iz_error  = 0   # or np.nan?

                concx   = c_rg_iz[0][1][i]
                rgx_, rgx_error   = c_rg_iz[1][1][i]


                rgx_errors.append( rgx_error )
                # atsas_rgx_errors.append( atsas_Rg_error )
                atsas_rgx_errors.append( 0 )        # suppress showing errorbars for atsas
                izx_, izx_error = c_rg_iz[2][1][i]

                rg_min = get_min( rg_min, rgx_, rgx_error )
                rg_max = get_max( rg_max, rgx_, rgx_error )
                rg_min = get_min( rg_min, atsas_Rg, atsas_Rg_error, atsas=True )
                rg_max = get_max( rg_max, atsas_Rg, atsas_Rg_error, atsas=True )
                iz_min = get_min( iz_min, izx_, izx_error )
                iz_max = get_max( iz_max, izx_, izx_error )
                iz_min = get_min( iz_min, atsas_Iz, atsas_Iz_error, atsas=True )
                iz_max = get_max( iz_max, atsas_Iz, atsas_Iz_error, atsas=True )
                # self.logger.info( 'debug: ' + str( (i, rg_min, rg_max, iz_min, iz_max) ) )

                # we have changed so that iz has been already scaled by /= concx
                # izx_errors.append( izx_error/concx if izx_error is not None and concx > 0 else 0 )
                # iz_concx = izx_ / concx if izx_ is not None and concx > 0 else None
                izx_errors.append( izx_error )
                iz_concx = izx_
                atsas_izx_errors.append( atsas_Iz_error )

                if i < zx_only_rows:
                    rg_errors.append( 0 )   # or np.nan?
                    iz_errors.append( 0 )   # or np.nan?
                    conc = 0 if i == 0 else None
                    ws.append( [ None, conc, concx,
                                    None, rgx_, atsas_Rg,
                                    None, iz_concx, atsas_Iz,
                                    None, rgx_error, atsas_Rg_error,
                                    None, izx_error, atsas_Iz_error ] )
                else:
                    nxi = i - zx_only_rows
                    rg_, rg_error, atsas_Rg, atsas_Rg_error = c_rg_iz[1][0][nxi]
                    rg_errors.append( rg_error )
                    iz_, iz_error, atsas_Iz, atsas_Iz_error = c_rg_iz[2][0][nxi]
                    # iz_ has not been scaled by /= concx
                    iz_error_c = iz_error/concx if iz_error is not None and concx > 0 else 0
                    iz_errors.append( iz_error_c )
                    iz_conc = iz_ / concx if iz_ is not None and concx > 0 else None
                    if atsas_Iz is not None:
                        atsas_Iz /= concx
                        atsas_Iz_error /= concx
                    ws.append( [    indeces[nxi],
                                    c_rg_iz[0][0][nxi], concx,
                                    rg_, rgx_, atsas_Rg,
                                    iz_conc, iz_concx, atsas_Iz,
                                    rg_error, rgx_error, atsas_Rg_error,
                                    iz_error_c, izx_error, atsas_Iz_error ] )
                    rg_min = get_min( rg_min, rg_, rg_error )
                    rg_max = get_max( rg_max, rg_, rg_error )
                    iz_min = get_min( iz_min, iz_conc, iz_error_c )
                    iz_max = get_max( iz_max, iz_conc, iz_error_c )

                # self.logger.info( 'debug: ' + str( (i, rg_min, rg_max, iz_min, iz_max) ) )
        else:
            for i in range(zx_rows):
                if atsas_result is None or i < zx_rows-1:
                    atsas_Rg        = None
                    atsas_Iz        = None
                else:
                    atsas_Rg        = atsas_result.Rg
                    # atsas_Rg_error  = atsas_result.Rg_stdev
                    # atsas_Iz should have been scaled by the max concentration
                    atsas_Iz        = divide( atsas_result.I0, self.c_max, None )
                    # atsas_Iz_error  = divide( atsas_result.I0_stdev, self.c_max, 0 )

                atsas_Rg_error  = 0   # or np.nan?
                atsas_Iz_error  = 0   # or np.nan?

                concx   = c_rg_iz[0][1][i]
                rgx_, rgx_error = c_rg_iz[1][1][i]
                rgx_errors.append( rgx_error )
                # atsas_rgx_errors.append( atsas_Rg_error )
                atsas_rgx_errors.append( 0 )        # suppress showing errorbars for atsas
                izx_, izx_error = c_rg_iz[2][1][i]

                rg_min = get_min( rg_min, rgx_, rgx_error )
                rg_max = get_max( rg_max, rgx_, rgx_error )
                rg_min = get_min( rg_min, atsas_Rg, atsas_Rg_error, atsas=True )
                rg_max = get_max( rg_max, atsas_Rg, atsas_Rg_error, atsas=True )
                iz_min = get_min( iz_min, izx_, izx_error )
                iz_max = get_max( iz_max, izx_, izx_error )
                iz_min = get_min( iz_min, atsas_Iz, atsas_Iz_error, atsas=True )
                iz_max = get_max( iz_max, atsas_Iz, atsas_Iz_error, atsas=True )
                # self.logger.info( 'debug: ' + str( (i, rg_min, rg_max, iz_min, iz_max) ) )

                # we have changed so that izx has been already scaled by /= concx
                # izx_errors.append( izx_error/concx if izx_error is not None and  concx > 0 else 0 )
                # iz_concx = izx_ / concx if izx_ is not None and concx > 0 else None
                izx_errors.append( izx_error )
                iz_concx = izx_
                atsas_izx_errors.append( atsas_Iz_error )

                if i < nx_rows:
                    rg_, rg_error, atsas_Rg, atsas_Rg_error = c_rg_iz[1][0][i]
                    rg_errors.append( rg_error )
                    iz_, iz_error, atsas_Iz, atsas_Iz_error = c_rg_iz[2][0][i]
                    # iz_ has not been scaled by /= concx
                    iz_error_c = iz_error/concx if iz_error is not None and  concx > 0 else 0
                    iz_errors.append( iz_error_c )
                    iz_conc = iz_ / concx if iz_ is not None and concx > 0 else None
                    if atsas_Iz is not None:
                        atsas_Iz /= concx
                        atsas_Iz_error /= concx
                    ws.append( [    indeces[i],
                                    c_rg_iz[0][0][i], concx,
                                    rg_, rgx_, atsas_Rg,
                                    iz_conc, iz_concx, atsas_Iz,
                                    rg_error, rgx_error, atsas_Rg_error,
                                    iz_error_c, izx_error, atsas_Iz_error ] )
                    rg_min = get_min( rg_min, rg_, rg_error )
                    rg_max = get_max( rg_max, rg_, rg_error )
                    iz_min = get_min( iz_min, iz_conc, iz_error_c )
                    iz_max = get_max( iz_max, iz_conc, iz_error_c )
                else:
                    rg_errors.append( 0 )   # or np.nan?
                    iz_errors.append( 0 )   # or np.nan?
                    ws.append( [ None, None, concx,
                                    None, rgx_, atsas_Rg,
                                    None, iz_concx, atsas_Iz,
                                    None, rgx_error, atsas_Rg_error, 
                                    None, izx_error, atsas_Iz_error ] )

                # self.logger.info( 'debug: ' + str( (i, rg_min, rg_max, iz_min, iz_max) ) )

        q_colname   = [ 'Q' ]

        colnames_ = q_colname

        if orig_param_array is None:
            colnames_ += [ 'A(q)', 'ATSAS(scaled)', 'B(q)' ]
        else:
            colnames_ += [ 'original A(q)', 'A(q)', 'ATSAS(scaled)', 'original B(q)', 'B(q)' ]
        ws.append( colnames_ )

        if almerge_result is not None:
            comp_atsas_Aq_list = []

        for i in range( num_intensity_points ):
            if exz_array is None:
                atsas_val = None
            else:
                atsas_val = exz_array[i,1] / self.c_max
                # A(q) intensity values should be positive for logarithmic scaling.
                # Negative values result in COM-drived Excel errors.
                if atsas_val <= 0:
                    atsas_val = None

            # A(q) intensity values should be positive for logarithmic scaling.
            # Negative values result in COM-drived Excel errors.
            a_ = param_array[i,0]
            if a_ <= 0:
                a_ = None

            if orig_param_array is None:
                param_array_slice =  np.array( [ a_, atsas_val, param_array[i,1] ] )
            else:
                # A(q) intensity values should be positive for logarithmic scaling.
                # Negative values result in COM-drived Excel errors.
                oa_ = orig_param_array[i,0]
                if oa_ <= 0:
                    oa_ = None
                param_array_slice = np.array( [ oa_, a_, atsas_val,
                                                orig_param_array[i,1], param_array[i,1],
                                              ] )
            ws.append( list( np.hstack( [ qvector[i], param_array_slice ] ) ) )
            if almerge_result is not None:
                a_ = np.nan if a_ is None else a_
                atsas_val = np.nan if atsas_val is None else atsas_val
                comp_atsas_Aq_list.append( [ a_, atsas_val ] )

        if almerge_result is not None:
            # TODO: disable this in product release
            self.compute_diff_atsas( comp_atsas_Aq_list, qvector )

        rg_errors = [ np.array( rg_errors ), np.array( rgx_errors ), np.array( atsas_rgx_errors ) ]
        iz_errors = [ np.array( iz_errors ), np.array( izx_errors ), np.array( atsas_izx_errors ) ]

        chart_start_col = 'Q'
        self.create_concetration_chart( ws, chart_start_col+'2' )
        self.create_Rg_chart( ws, 'AB2', rg_errors, [rg_min, rg_max] )
        self.create_Iz_conc_chart( ws, 'AM2', iz_errors, [iz_min, iz_max] )
        self.create_extrapolation_params_chart( ws, zx_rows+1, 2, 0, PARAM_A_NAME, chart_start_col+'33', r_adjust=-0.02 )
        if self.need_bq:
            self.create_extrapolation_params_chart( ws, zx_rows+1, 2, 1, PARAM_B_NAME, chart_start_col+'85' )

    def compute_diff_atsas( self, comp_atsas_Aq_list, qvector ):
        if self.boundary_j is None:
            return

        j = self.boundary_j
        comp_atsas_array = np.array( comp_atsas_Aq_list[0:j] )
        k_val = comp_atsas_array[:,0]
        a_val = comp_atsas_array[:,1]
        isfinite = np.logical_and( np.isfinite(k_val), np.isfinite(a_val) )
        try:
            diff_atsas = np.sum( np.abs( (k_val - a_val)[isfinite] ) )
            max_val = np.max( a_val[isfinite] )
        except:
            diff_atsas = None
            max_val = None
        write_to_tester_log( 'ex-curves-diff=' + str( [ diff_atsas, max_val, j, qvector[j] ] ) + '\n' )

    def create_params_chart( self, ws, pos, title, y_title, x_title, y_cols, x_cols, errors=None, minmax=None ):
        f, t = y_cols
        if t-f == 1:
            c_ = LineChart()
        else:
            c_ = ScatterChart()
        c_.title = title
        c_.style = 13
        c_.y_axis.title = y_title
        c_.x_axis.title = x_title

        xvalues0 = Reference(ws, min_col=x_cols[0],
                                min_row=1 + 1,      # must not include column title
                                max_row=self.zx_rows+1)

        if t-f == 1:
            data = Reference(ws, min_col=f, max_col=t, min_row=1, max_row=self.zx_rows+1 )
            c_.add_data(data, titles_from_data=True)
            c_.set_categories(xvalues0)
            c_.width    = CHART_WIDTH_ZX_P
            c_.height   = CHART_HEIGHT_ZX_P
        else:
            assert t-f == 2
            xvalues1 = Reference(ws, min_col=x_cols[1],
                                    min_row=1 + 1,      # must not include column title
                                    max_row=self.zx_rows+1)
            for col in [f, f+1, t]:
                data = Reference(ws, min_col=col, min_row=1, max_row=self.zx_rows+1 )
                xvalues_ = xvalues1 if col == f+1 else xvalues0
                series  = Series(data, xvalues_, title_from_data=True)
                c_.series.append(series)
                c_.width    = CHART_WIDTH_ZX_P
                c_.height   = CHART_HEIGHT_ZX_P

        if minmax is not None:
            # self.logger.info( 'minmax=' + str(minmax) )
            min_, max_ = minmax
            # TODO: improve scale
            if max_ is None:
                # max_ is None for 2160227. why?
                self.logger.warning( 'Unexpected state: minmax values seem to be None in create_params_chart' )
            else:
                if max_ > 10:
                    scale = 1
                else:
                    scale = 1e-3
                c_.y_axis.scaling.min   = np.floor( min_/scale )*scale
                c_.y_axis.scaling.max   = np.ceil( max_/scale )*scale

        colors = [ green, orange ]
        if t - f == 2:
            # i.e., for Rg and Iz_con
            # TODO: better expression
            colors.append( ATSAS_COLOR )

        # print( 'errors=', errors )

        # Style the lines

        for i, color in enumerate(colors):
            s_ = c_.series[i]

            if ADD_ERROR_BARS and errors:
                plus    = errors[i]
                minus   = plus
                errorbars = list2errorbars(plus, minus, errDir='y')
                s_.errBars = errorbars

            s_.marker.symbol = "circle"
            s_.marker.graphicalProperties.solidFill         = color
            s_.marker.graphicalProperties.line.solidFill    = color
            s_.graphicalProperties.line.solidFill           = color
            s_.graphicalProperties.line.width               = LINE_WIDTH    # in EMUs

        set_layout_small( c_ )

        ws.add_chart(c_, pos)

    def create_concetration_chart( self, ws, pos ):
        self.create_params_chart( ws, pos, 'Concentration', 'Concentration', 'Elution №', [ 2, 3 ], [ 1, 1 ] )

    def create_Rg_chart( self, ws, pos, errors, minmax ):
        self.create_params_chart( ws, pos, 'Rg Extrapolated', 'Rg', 'Concentration', [ 4, 6 ], [ 2, 3 ], errors=errors, minmax=minmax )

    def create_Iz_conc_chart( self, ws, pos, errors, minmax ):
        self.create_params_chart( ws, pos, 'I(0)/Conc. Extrapolated', 'I(0)/Conc.', 'Concentration', [ 7, 9 ], [ 2, 3 ], errors=errors, minmax=minmax )

    def create_extrapolated_intensity_chart( self, ws, row_offset, col_offset, pos):
        c_ = ScatterChart()
        c_.title = "Extrapolation to Zero Concentration"
        c_.style = 13
        c_.y_axis.title = 'Intensity / Concentration' if col_offset == 0 else 'Intensity'
        c_.x_axis.title = 'Q(Å⁻¹)'

        min_row = row_offset + 1
        max_row = row_offset + self.num_rows + 1

        xvalues = Reference(ws, min_col=1,
                                min_row=min_row + 1,    # must not include column title
                                max_row=max_row)

        for col in range(col_offset+2, col_offset+self.nx_rows+self.zx_rows+2):
            values  = Reference(ws, min_col=col, min_row=min_row, max_row=max_row)
            series  = Series(values, xvalues, title_from_data=True)
            c_.series.append(series)

        c_.width    = CHART_WIDTH_ZX_I
        c_.height   = CHART_HEIGHT_ZX_I

        orange_list = [ orange ] * ( self.zx_rows - 1 )
        colors_x = ( [ red ] + orange_list ) if self.ud == 0 else ( orange_list + [ red ] )
        colors = [ green ] * self.nx_rows + colors_x
        # Style the lines

        for i, color in enumerate(colors):
            s_ = c_.series[i]
            # s_.marker.symbol = "circle"
            s_.marker.graphicalProperties.solidFill         = color
            s_.marker.graphicalProperties.line.solidFill    = color
            s_.graphicalProperties.line.solidFill           = color
            width = LINE_WIDTH    # in EMUs
            if color == red:
                width *= 2
            s_.graphicalProperties.line.width               = width

        set_layout_large( c_ )

        ws.add_chart(c_, pos)

    def create_extrapolation_params_chart( self, ws, row_offset, col_offset, j, param_name, pos, r_adjust=0):

        atsas_col = 0 if ( self.exz_array is None or j == 1 ) else 1
        no_orig_param_array = self.orig_param_array is None

        chart = create_extrapolation_params_chart_( ws, self.ud, self.num_rows, row_offset, col_offset, j, param_name, atsas_col, no_orig_param_array, r_adjust )

        ws.add_chart(chart, pos)

    def add_format_setting( self, book_file ):
        from .ZeroExExcelFormatter import ZeroExReportArgs
        axis_direction_desc = get_setting( 'axis_direction_desc' )
        args = ZeroExReportArgs(self.ws.title, self, book_file, axis_direction_desc)    # make it picklable

        if self.parent.more_multicore:
            self.add_format_setting_more_multicore(args)
        else:
            self.add_format_setting_less_multicore(args)

    def add_format_setting_more_multicore(self, args):
        self.parent.teller.tell('range_extrapolation_book', args=args)

    def add_format_setting_less_multicore(self, args):
        from .ZeroExExcelFormatter import add_result_format_setting

        add_result_format_setting(self.parent.excel_client, args, self.logger)

    def save( self, xlsx_file ):
        save_allowing_user_reply( self.wb, xlsx_file )



"""
    G0525.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy.Peaks.ElutionModels import egh
from molass_legacy.Optimizer.BasicOptimizer import BasicOptimizer, PENALTY_SCALE
from molass_legacy.Optimizer.NumericalUtils import safe_ratios
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.ModelParams.SeccolFunctions import rgfit_secconf_eval
from molass_legacy.SecTheory.BoundControl import Penalties
from molass_legacy.Optimizer.TheDebugUtils import convert_score_list

UV_B_ALLOW_RATIO = 0.1
EGH_LOG_ALPHA = np.log(0.1)
UV_XR_RATIO_ALLOW = 0.25
BAD_PARAMS_RETURN = 1e8

class G0525(BasicOptimizer):
    """
    Constrained EGH (Lan-Jorgenson)
    """
    def __init__(self, dsets, n_components, **kwargs):

        t0_upper_bound = get_setting("t0_upper_bound")
        if t0_upper_bound is None:
            from molass_legacy.SecTheory.T0UpperBound import estimate_t0upper_bound
            ecurve = dsets[0][0]
            t0_upper_bound = estimate_t0upper_bound(ecurve)
            set_setting("t0_upper_bound", t0_upper_bound)

        if True:
            from importlib import reload
            import molass_legacy.ModelParams.LjEghParams
            reload(molass_legacy.ModelParams.LjEghParams)
        from molass_legacy.ModelParams.LjEghParams import LjEghParams

        params_type = LjEghParams(n_components)

        BasicOptimizer.__init__(self, dsets, n_components, params_type, kwargs)

    def update_bounds(self, x, debug=False):
        self.sigma_bounds = self.params_type.get_sigma_bounds(x)
        self.logger.info("sigma_bounds have been estimated as %s", str(self.sigma_bounds))
        super().update_bounds(x, debug=debug)

    def objective_func(self, p, plot=False, debug=False, fig_info=None, axis_info=None, return_full=False, return_lrf_info=False):
        self.eval_counter += 1
        try:
            xr_params, xr_baseparams, rg_params, (a, b), uv_params, uv_baseparams, (c, d), seccol_params = self.split_params_simple(p)
            Npc, tI = seccol_params[[0,2]]
        except ValueError:
            return BAD_PARAMS_RETURN

        x = self.xr_curve.x
        y = self.xr_curve.y
        # rg = self.rg

        uv_x = a*x+b
        uv_y = self.uv_curve.spline(uv_x)

        a_init, b_init = self.init_mapping
        ratio = a/a_init
        a_deviation = min(0, ratio - 0.99)**2 + max(0, ratio - 1.05)**2
        b_allowance = len(self.uv_curve.x) * UV_B_ALLOW_RATIO
        b_deviation = max(0, abs(b - b_init) - b_allowance)**2

        uv_xr_ratio = uv_params/xr_params[:,0]
        ratio_deviation = max(0, np.std(uv_xr_ratio)/np.average(uv_xr_ratio) - UV_XR_RATIO_ALLOW)**2
        if debug:
            self.logger.info("uv_xr_ratio=%s, ratio_deviation=%.3g", str(uv_xr_ratio), ratio_deviation)

        mapping_penalty = PENALTY_SCALE * (a_deviation + b_deviation ) + ratio_deviation        # ratio_deviation should not be scaled to PENALTY_SCALE

        xr_cy_list = []
        uv_cy_list = []
        xr_ty = np.zeros(len(x))
        uv_ty = np.zeros(len(uv_x))
        intercept_penalty = 0
        baseline_penalty = 0
        masked_params = p[self.bounds_mask]
        outofbounds_penalty = PENALTY_SCALE * (np.sum(np.max([self.zero_bounds, self.lower_bounds - masked_params], axis=0)) + np.sum(np.max([self.zero_bounds, masked_params - self.upper_bounds], axis=0)) + Penalties[0])
        if self.eval_counter == 1 and outofbounds_penalty > 0:
            self.logger.info("out of lower bounds: %s", str(np.max([self.zero_bounds, self.lower_bounds - masked_params], axis=0)))
            self.logger.info("out of upper bounds: %s", str(np.max([self.zero_bounds, masked_params - self.upper_bounds], axis=0)))
        if plot:
            overlap = np.zeros(len(x))
        last_mu = None
        last_cy = None
        sec_penality = 0

        average_m = np.average(xr_params[:,1])
        model_s = (average_m - tI)/np.sqrt(Npc)
        outofbounds_penalty += PENALTY_SCALE * (min(0, model_s - self.sigma_bounds[0])**2 + max(0, model_s - self.sigma_bounds[1])**2)

        order_penalty = 0
        k = 0
        for h, m, s, t in xr_params:
            xr_cy = egh(x, h, m, s, t)
            uv_cy = egh(uv_x, uv_params[k], a*m+b, a*s, a*t)

            if k > 0:
                order_penalty += max(0, last_mu - m)**2
                if debug:
                    overlap += np.abs(np.min([last_cy, xr_cy], axis=0))     # np.abs() is intended to degrade negative elements

            last_mu = m
            last_cy = xr_cy

            xr_ty += xr_cy
            xr_cy_list.append(xr_cy)
            uv_ty += uv_cy
            uv_cy_list.append(uv_cy)
            k += 1

        xr_cy = self.xr_baseline(x, xr_baseparams, xr_ty, xr_cy_list)
        uv_cy = self.uv_baseline(uv_x, uv_baseparams, uv_ty, uv_cy_list)
        xr_ty += xr_cy
        xr_cy_list.append(xr_cy)
        uv_ty += uv_cy
        uv_cy_list.append(uv_cy)

        try:
            lrf_info = self.compute_LRF_matrices(x, y, xr_cy_list, xr_ty, uv_x, uv_y, uv_cy_list, uv_ty)
            if return_lrf_info:
                return lrf_info

            m, s = xr_baseparams[0:2]   # how about r?

            slope_penalty = max(self.slope_allowance, (m - self.init_slope)**2) - self.slope_allowance
            intercept_penalty = max(self.intercept_allowance, (s - self.init_intercept)**2) - self.intercept_allowance
            baseline_penalty += slope_penalty*self.slope_penalty_scale + intercept_penalty*self.intercept_penalt_scale
            if self.avoid_peak_fronting:
                negative_penalty = PENALTY_SCALE * (  min(0, np.min(xr_params[:,0]))**2     # h
                                                    + min(0, np.min(xr_params[:,3]))**2     # tau
                                                    + min(0, np.min(uv_params[:]))**2       # h
                                                    + min(0, np.min(rg_params))**2          # rg
                                                    + intercept_penalty)
            else:
                negative_penalty = PENALTY_SCALE * (  min(0, np.min(xr_params[:,0]))**2     # h
                                                    + min(0, np.min(uv_params[:]))**2       # h
                                                    + min(0, np.min(rg_params))**2          # rg
                                                    + intercept_penalty)

            order_penalty *= PENALTY_SCALE
            penalties = [mapping_penalty, negative_penalty, baseline_penalty, outofbounds_penalty, order_penalty]

            fv, score_list = self.compute_fv(lrf_info, xr_params, rg_params, seccol_params, penalties, p, debug=debug)
        except:
            # e.g., numpy.linalg.LinAlgError: SVD did not converge
            log_exception(self.logger, "error in objective_func", n=5)
            fv = np.inf
            score_list = [0] * self.get_num_scores([])
            matrices = [None] * self.get_num_matrices()
            xr_ty = np.zeros(len(y))
            if False:
                debug_ = True
                overlap = np.zeros(len(x))

        if plot:
            from molass_legacy.ModelParams.EghPlotUtils import plot_objective_state
            debug_fv = plot_objective_state((score_list, penalties), fv,
                lrf_info,
                overlap, self.rg_curve, rg_params,
                self.get_score_names(),
                fig_info, axis_info,
                self, p,
                )
            if axis_info is None:
                self.debug_fv = debug_fv

        if return_full:
            score_list = convert_score_list((score_list, penalties))
            return fv, score_list, *lrf_info.matrices
        else:
            return fv



"""
    Console.py
"""
import sys
import time
import win32console, win32con
import psutil

class Console:
    def __init__( self, windows_app_only=False, console_parent_only=False ):
        self.activeted = False

        proc    = psutil.Process()
        pproc   = psutil.Process(proc.ppid())
        ppproc  = psutil.Process(pproc.ppid())
        pname   = pproc.name()
        ppname  = ppproc.name()
        parent_is_console = 'cmd.exe' in [ pname, ppname ]
        if console_parent_only and not parent_is_console:
            return

        free_console=True
        try:
            win32console.AllocConsole()
        except win32console.error as exc:
            if exc.winerror!=5:
                raise
            ## only free console if one was created successfully
            free_console=False

        if windows_app_only:
            if not free_console: return

        self.activeted = True

        self.free_console = free_console
        self.stdout     = win32console.GetStdHandle(win32console.STD_OUTPUT_HANDLE)
        self.stdin      = win32console.GetStdHandle(win32console.STD_INPUT_HANDLE)

        self.newbuffer  = newbuffer = win32console.CreateConsoleScreenBuffer()

        newbuffer.SetConsoleActiveScreenBuffer()
        """
        newbuffer.SetConsoleTextAttribute(win32console.FOREGROUND_RED|win32console.FOREGROUND_INTENSITY
                |win32console.BACKGROUND_GREEN|win32console.BACKGROUND_INTENSITY)
        """
        newbuffer.WriteConsole('This is a new screen buffer\n')

        ## test setting screen buffer and window size
        ## screen buffer size cannot be smaller than window size
        window_size=newbuffer.GetConsoleScreenBufferInfo()['Window']
        coord=win32console.PyCOORDType(X=window_size.Right+20, Y=window_size.Bottom+20)
        newbuffer.SetConsoleScreenBufferSize(coord)

        window_size.Right+=10
        window_size.Bottom+=10

        newbuffer.SetConsoleWindowInfo(Absolute=True,ConsoleWindow=window_size)
        self.stdout_save = sys.stdout
        sys.stdout      = self

        # self.write( 'free_console=' + str( self.free_console ) + '\n' )

    def __del__( self ):
        if not self.activeted: return

        print( 'press <Control> to close.' )

        breakout=False
        while not breakout:
            input_records = self.stdin.ReadConsoleInput(1)
            for input_record in input_records:
                if input_record.VirtualKeyCode==win32con.VK_END:
                    breakout=True

            time.sleep(0.1)

        self.stdout.SetConsoleActiveScreenBuffer()

        if self.free_console:
             win32console.FreeConsole()
        sys.stdout = self.stdout_save

    def write( self, buffer ):
        self.newbuffer.WriteConsole( buffer )

    def flush( self, *args ):
        pass

"""
    https://stackoverflow.com/questions/287871/print-in-terminal-with-colors
    http://vivi.dyndns.org/tech/cpp/setColor.html
"""
import ctypes

STD_OUTPUT_HANDLE = -11
COLOR_WHITE = 0x0f
COLOR_GRAY  = 0x07

def reset_text_color():
    handle = ctypes.windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)
    ctypes.windll.kernel32.SetConsoleTextAttribute(handle, COLOR_GRAY)



"""
    ConcDepend.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF
"""
import logging
from bisect import bisect_right
import numpy as np
from scipy.optimize import minimize
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from molass_legacy._MOLASS.SerialSettings import get_xray_picking

LIMIT_Q = 0.2
CD_COLORS = ['green', 'yellow', 'red']
# CD_BOUNDARIES = [0.5, 1]
CD_BOUNDARIES = [2, 5]
BOUNDED_SCALING = True
USE_SMOOTHED_AQ_SCALED_BQ = True
RDR_COMPUTE_LIMIT_RATIO = 0.5
RANK2_RDR_LIMIT = 0.02  # < 0.02xx for 20190527 (EndoAB)
RANK2_SCD_BY_RDR = 3    # this makes the color to be 'yellow'

def compute_min_norm_scaled(y1, y2):
    def obj_func(p):
        return np.linalg.norm(y1-p[0]*y2)**2
    result = minimize(obj_func, np.ones(1))
    scale = result.x[0]
    return scale*y2

def compute_min_norm_scaled_wide(y1, y2, wslice):
    y1_ = y1[wslice]
    y2_ = y2[wslice]
    def obj_func(p):
        return np.linalg.norm(y1_-p[0]*y2_)**2
    result = minimize(obj_func, np.ones(1))
    scale = result.x[0]
    print('scale=', scale)
    return scale*y2

def compute_min_norm_bq_s_aq(y1, y2, init_scale, logger=None, ret_no_bnd=False, ret_scale=False):
    def obj_func(p):
        return np.linalg.norm(y1-p[0]*y2)**2

    # bnd = init_scale*2
    # result = minimize(obj_func, (init_scale,), bounds=((-bnd, +bnd),))
    result = minimize(obj_func, (init_scale,))
    scale = result.x[0]
    if BOUNDED_SCALING:
        ratio = abs(scale/init_scale)
        if logger is not None:
            logger.info('scale ratio: %.3g/%.3g=%.3g', scale, init_scale, ratio)
        sy2_no_bnd = None
        if ratio > 3 and False:
            # as for 20180526/OA
            sy2_no_bnd = scale*y2
            scale = 0
            if logger is not None:
                logger.info('scale is set to zero due to too large ratio %.3g', ratio)
        if ret_no_bnd:
            return scale*y2, sy2_no_bnd

    if ret_scale:
        return scale*y2, scale
    else:
        return scale*y2

def compute_distinct_cd(M, C, cd_slice, scale, logger):
    from molass_legacy.QuickAnalysis.JudgeHolder import convert_to_degrees
    try:
        U, sigmas, VT = np.linalg.svd(M)
        rank = 2
        M_ = U[:,0:rank] @ np.diag(sigmas[0:rank]) @ VT[0:rank,:]

        c = C[0,:]
        c = c/np.max(c)
        C_ = np.array([c, c**2])
        P = M_ @ np.linalg.pinv(C_)
        a = smooth(P[:,0])
        b = smooth(P[:,1])

        init_scale = sigmas[1]/sigmas[0]
        sa = compute_min_norm_bq_s_aq(b, a, init_scale, logger)
        y = b - sa
        cd_score = np.sqrt(np.average(y[cd_slice]**2))*100/scale
        cd = convert_to_degrees([cd_score])

        logger.info("cd=%d from cd_score:%.3g", cd[0], cd_score)
        ret_cd = cd[0]
    except:
        # as in 20201122
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(logger, "compute_distinct_cd failed: ")
        ret_cd = 1
        logger.warning("cd=%d due to the last error.", ret_cd)

    return ret_cd

class ConcDepend:
    def __init__(self, q, data, error, x_curve, xray_scale=None):
        self.logger = logging.getLogger(__name__)
        self.q = q
        self.xray_picking = get_xray_picking()
        self.i = bisect_right(q, self.xray_picking)
        self.ip = bisect_right(q, self.xray_picking+0.01)
        self.limit = bisect_right(q, LIMIT_Q)
        self.data = data
        self.error = error
        self.x_curve = x_curve
        self.xray_scale = xray_scale
        self.rdr_hints = None

    def compute_judge_info(self, rdr_hints, debug_info=None):
        x_curve = self.x_curve
        aslice = slice(0, self.limit)
        scd_list = []
        rdr_list = []
        islice = slice(self.i, self.ip)
        x = self.x_curve.x
        max_y = self.x_curve.max_y

        if len(x_curve.peak_info) == len(rdr_hints):
            # as in most usual cases
            proc_peak_info = x_curve.peak_info
        else:
            # as in 20170506/ACTM002/Backsub with an ecurve other than x_curve
            # this may have been fixed with x_curve, i.e., no need of this else clause
            proc_peak_info = x_curve.get_major_peak_info()

        if debug_info is not None:
            if len(rdr_hints) != len(proc_peak_info):
                # as in 
                from OnTheFly.DebugDialog import DebugDialog
                dialog = DebugDialog(debug_info=[rdr_hints, x_curve, debug_info[0]])
                dialog.show()

        for k, peak_rec in enumerate(proc_peak_info):
            print([k], peak_rec)
            start = peak_rec[0]
            stop = peak_rec[2]+1
            if start >= stop:
                # as in 20171226
                # or should be investigated for the cause
                continue

            eslice = slice(start, stop)
            M = self.data[:,eslice]
            E = self.error[:,eslice]
            U, s, VT= np.linalg.svd(M)
            print('s[0:2]=', s[0:2])
            M1_ = get_denoised_data(M, rank=1)
            M2_ = get_denoised_data(M, rank=2)
            c = x_curve.y[eslice]
            c = c/np.max(c)
            C1 = np.array([c])
            C2 = np.array([c, c**2])
            C1inv = np.linalg.pinv(C1)
            C2inv = np.linalg.pinv(C2)
            P1 = np.dot(M1_, C1inv)
            P2 = np.dot(M2_, C2inv)
            A1 = P1[:,0]
            A2 = P2[:,0]
            B2 = P2[:,1]

            if self.xray_scale is None:
                s1 = np.average(A1[islice])
                s2 = np.average(A2[islice])
                scale = np.sqrt(s1*s2)
            else:
                scale = self.xray_scale

            if USE_SMOOTHED_AQ_SCALED_BQ:
                init_scale = s[1]/s[0]
                B2_ = smooth(B2)
                A2_ = smooth(A2)
                sy2 = compute_min_norm_bq_s_aq(B2_, A2_, init_scale)
                y = B2_ - sy2
            else:
                sy2 = compute_min_norm_scaled(A2, B2)
                y = A2 - sy2
            scd = np.sqrt(np.average(y[aslice]**2))*100/scale
            # bqn = np.sqrt(np.average(B2[aslice]**2))*100/scale
            # print([k], 'scd, bqn=', scd, bqn)
            top_x = peak_rec[1]

            computable, rdr = rdr_hints[k]
            if not computable:
                # to avoid replacing the scd
                rdr = 0

            height_ratio = self.x_curve.spline(top_x)/max_y
            if height_ratio < RDR_COMPUTE_LIMIT_RATIO or scd >= CD_BOUNDARIES[0]:
                # keep the scd as is
                pass
            else:
                # rdr = self.compute_rdr_legacy(M, E, P1, A1, A2)
                if abs(rdr) > RANK2_RDR_LIMIT:
                    self.logger.info("[%d] scd %.3g has been changed to %.3g due to rdr %.3g", k, scd, RANK2_SCD_BY_RDR, rdr)
                    scd = RANK2_SCD_BY_RDR

            scd_list.append((top_x, scd))
        return scd_list

    def compute_rdr_legacy(self, M, E, P1, A1, A2):
        from molass_legacy.QuickAnalysis.RgDiffRatios import compute_rdr_legacy_impl
        return compute_rdr_legacy_impl(self.q, M, E, P1, A1, A2)



"""
    from
        Molecular Dynamics
        https://pythoninchemistry.org/sim_and_scat/molecular_dynamics/intro.html
"""
import numpy as np
import matplotlib.pyplot as plt

def lj_force(r, epsilon, sigma):
    """
    Implementation of the Lennard-Jones potential 
    to calculate the force of the interaction.
    
    Parameters
    ----------
    r: float
        Distance between two particles (Å)
    epsilon: float 
        Potential energy at the equilibrium bond 
        length (eV)
    sigma: float 
        Distance at which the potential energy is 
        zero (Å)
    
    Returns
    -------
    float
        Force of the van der Waals interaction (eV/Å)
    """
    return 48 * epsilon * np.power(
        sigma, 12) / np.power(
        r, 13) - 24 * epsilon * np.power(
        sigma, 6) / np.power(r, 7)
    
r = np.linspace(3.5, 8, 100)
plt.plot(r, lj_force(r, 0.0103, 3.4))
plt.xlabel(r'$r$/Å')
plt.ylabel(r'$f$/eVÅ$^{-1}$')
plt.show()



# -*- coding: utf-8 -*-
"""

    ファイル名：   Development.py

    処理内容：

        GUI のオプション設定情報

"""
from __future__ import division, print_function, unicode_literals

from PersistentInfo import PersistentInfo

DEFAULT_DEVELOPER_OPTIONS = {
    'adj_algorithm'         : 'round',
    'min_ratio'             : 0.5,
    'adj_output'            : 'NO',
    'postfix_adj'           : '_adj',
    'intermediate_results'  : 'NO',
    'defaultfont'           : ( 'TkDefaultFont', 10 ),
    'fixedfont'             : ( 'TkFixedFont', 10 ),
    'debug'                 : False,
    }

development_info    = PersistentInfo( 'development.dump', DEFAULT_DEVELOPER_OPTIONS )
development_        = development_info.get_dictionary()

from molass_legacy.KekLib.OurTkinter import Tk, Dialog, tk_set_icon_portable

DEBUG = False

def get_devel_info( item ):
    assert( item in DEFAULT_DEVELOPER_OPTIONS )
    return development_.get( item )

def set_devel_info( item, value ):
    assert( item in DEFAULT_DEVELOPER_OPTIONS )
    development_[item] = value

class DeveloperOptionsDialog( Dialog ):

    def __init__( self, parent, title ):
        self.grab = 'local'     # used in grab_set
        self.temporary          = False
        self.applied            = False

        Dialog.__init__(self, parent, title) # this calls body

    def body( self, body_frame ):   # overrides parent class method

        tk_set_icon_portable( self, 'synthesizer' )

        iframe = Tk.Frame( body_frame );
        iframe.pack( expand=1, fill=Tk.BOTH, padx=10, pady=10 )

        label_spacing = '     '
        text_entry_width    = 12

        grid_row = -1

        grid_row += 1
        adj_algorithm_label = Tk.Label( iframe, text= 'Adjust Algorithm:' + label_spacing )
        adj_algorithm_label.grid( row=grid_row, column=0, sticky=Tk.E )
        self.adj_algorithm = Tk.StringVar()
        j = 0
        for t, method in ( [ 'Round', 'round' ], [ 'Fast', 'fast'], [ 'Slow', 'slow' ]  ):
            text_ = '%-20.20s' % ( t )
            b = Tk.Radiobutton( iframe, text=text_,
                            variable=self.adj_algorithm, value=method )
            j += 1
            b.grid( row=grid_row, column=j, sticky=Tk.W )

        grid_row += 1
        self.min_ratio = Tk.DoubleVar()
        self.min_ratio.set( get_devel_info( 'min_ratio' ) )
        min_ratio_label = Tk.Label( iframe, text= 'Acceptable Pixel Cover Ratio:' + label_spacing )
        min_ratio_label.grid( row=grid_row, column=0, sticky=Tk.E )
        min_ratio_entry = Tk.Entry( iframe, textvariable=self.min_ratio, width=text_entry_width, justify=Tk.CENTER )
        min_ratio_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        def adj_algorithm_trace( *args ):
            adj_algorithm = self.adj_algorithm.get()
            if adj_algorithm == 'round':
                state_ = 'disabled'
            else:
                state_ = 'normal'
            min_ratio_entry.config( state=state_ )

        self.adj_algorithm.trace( "w", adj_algorithm_trace )
        self.adj_algorithm.set( get_devel_info( 'adj_algorithm' ) )

        grid_row += 1
        adj_output_label = Tk.Label( iframe, text= 'Adjusted File Output:' + label_spacing )
        adj_output_label.grid( row=grid_row, column=0, sticky=Tk.E )
        self.adj_output = Tk.StringVar()
        j = 0
        for t, n in ( [ 'Yes', 'YES' ], [ 'No', 'NO' ] ):
            b = Tk.Radiobutton( iframe, text=t,
                            variable=self.adj_output, value=n )
            j += 1
            b.grid( row=grid_row, column=j, sticky=Tk.W )

        grid_row += 1
        self.postfix_adj = Tk.StringVar()
        self.postfix_adj.set( get_devel_info( 'postfix_adj' ) )
        postfix_adj_label = Tk.Label( iframe, text= 'Adjusted File Postfix:' + label_spacing )
        postfix_adj_label.grid( row=grid_row, column=0, sticky=Tk.E )
        postfix_adj_entry = Tk.Entry( iframe, textvariable=self.postfix_adj, width=text_entry_width, justify=Tk.CENTER )
        postfix_adj_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        def adj_output_trace( *args ):
            adj_output = self.adj_output.get()
            if adj_output == 'NO':
                state_ = 'disabled'
            else:
                state_ = 'normal'
            postfix_adj_entry.config( state=state_ )

        self.adj_output.trace( "w", adj_output_trace )

        self.adj_output.set( get_devel_info( 'adj_output' ) )

        grid_row += 1
        intermediate_results_label = Tk.Label( iframe, text= 'Intermediate Results Output :' + label_spacing )
        intermediate_results_label.grid( row=grid_row, column=0, sticky=Tk.E )
        self.intermediate_results = Tk.StringVar()
        self.intermediate_results.set( get_devel_info( 'intermediate_results' ) )
        j = 0
        for t, n in ( [ 'Yes', 'YES' ], [ 'No', 'NO' ] ):
            b = Tk.Radiobutton( iframe, text=t,
                            variable=self.intermediate_results, value=n )
            j += 1
            b.grid( row=grid_row, column=j, sticky=Tk.W )

    def select_font( self ):
        font = askChooseFont( self )
        print( 'selected font=', font )
        pass

    def apply( self ):  # overrides parent class method
        print( "ok. apply" )

        set_devel_info( 'min_ratio',            self.min_ratio.get() )
        set_devel_info( 'adj_algorithm',        self.adj_algorithm.get() )
        set_devel_info( 'intermediate_results', self.intermediate_results.get() )

        old_adj_output = get_devel_info( 'adj_output' )
        new_adj_output = self.adj_output.get()
        if new_adj_output != old_adj_output:
            self.adj_output_changed = True
        set_devel_info( 'adj_output', new_adj_output )
        set_devel_info( 'postfix_adj', self.postfix_adj.get() )

        if not self.temporary:
            # development_info.save()
            pass    # don't save

        self.applied    = True



"""
    Models.Stochastic.PaperKnox1984.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt

def data_for_cylinder_along_z(center_x,center_y,radius,height_z):
    """
    https://stackoverflow.com/questions/26989131/add-cylinder-to-plot
    """
    z = np.linspace(0, height_z, 50)
    theta = np.linspace(0, 2*np.pi, 50)
    theta_grid, z_grid=np.meshgrid(theta, z)
    x_grid = radius*np.cos(theta_grid) + center_x
    y_grid = radius*np.sin(theta_grid) + center_y
    return x_grid,y_grid,z_grid

def illust_cylindrical():
    from Theory.SpherePlot import sphere_xyz
    fig, axes = plt.subplots(ncols=3, figsize=(15,5),  subplot_kw=dict(projection='3d'))
    fig.suptitle("Illustation of $K_{SEC}$ Estimation in Cylindrical Pores based on J. H. Knox, 1984", fontsize=20)

    R = 0.5

    for ax, r in zip(axes, [0.1, 0.2, 0.3]):
        rho = r/R
        Ksec = (1 - rho)**2
        ax.set_title("R=%g, r=%g\n\n" % (R, r) + r"$K_{SEC} = (\frac{R - r}{R})^2 = (1 - \frac{r}{R})^2 = (1 - \rho)^2$ = %.3g" % Ksec, y=0.95)
        ax.set_axis_off()

        Xc,Yc,Zc = data_for_cylinder_along_z(0.5, 0.5, R, 1)
        ax.plot_surface(Xc, Zc, Yc, alpha=0.3, label="Pore: Radius=R=%.3g" % R)

        Xc,Yc,Zc = data_for_cylinder_along_z(0.5, 0.5, R - r, 1)
        ax.plot_surface(Xc, Zc, Yc, alpha=0.3, label="Particle Center Boundary: Radius=R - r=%.3g" % (R - r))

        x, y, z = sphere_xyz(center=(0.5, 0.0, r), r=r)
        ax.plot_surface(x, y, z, facecolor='C2', alpha=0.3, label="Particles: Radius=r=%.3g" % r)

        x, y, z = sphere_xyz(center=(r, 0.0, 0.5), r=r)
        ax.plot_surface(x, y, z, facecolor='C2', alpha=0.3)

        x, y, z = sphere_xyz(center=(0.5, 0.0, 1-r), r=r)
        ax.plot_surface(x, y, z, facecolor='C2', alpha=0.3)

        ax.legend(loc="lower right")
        ax.view_init(elev=10, azim=-70, roll=0)

    fig.tight_layout()
    plt.show()

def sphere_xyz_alt(center=(0, 0, 0), radius=1, n=100, phi=None, theta=None):
    if phi is None:
        phi = np.linspace(0, np.pi, 100)
    if theta is None:
        theta = np.linspace(0, 2 * np.pi, 100)

    phi, theta = np.meshgrid(phi, theta)
    x = center[0] + radius * np.sin(phi) * np.cos(theta)
    y = center[1] + radius * np.sin(phi) * np.sin(theta)
    z = center[2] + radius * np.cos(phi)
    return x, y, z

def test_sphere_xyz_alt():
    n = 50
    phi = np.linspace(np.pi*0.2, np.pi*0.8, n)
    x, y, z = sphere_xyz_alt(n=n, phi=phi)
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d'))
    ax.set_axis_off()
    ax.plot_surface(x, z, y, alpha=0.3)
    ax.view_init(elev=10, azim=-70, roll=0)
    fig.tight_layout()
    plt.show()

def illust_spherical():
    from Theory.SpherePlot import sphere_xyz
    fig, axes = plt.subplots(ncols=3, figsize=(15,5),  subplot_kw=dict(projection='3d'))
    fig.suptitle("Illustation of $K_{SEC}$ Estimation in Spherical Pores based on J. H. Knox, 1984", fontsize=20)


    R = 0.5
    r = 0.2
    n = 50
    # phi = np.linspace(np.pi*0.2, np.pi*0.8, n)
    phi = None

    for ax, r in zip(axes, [0.1, 0.2, 0.3]):
        rho = r/R
        Ksec = (1 - rho)**3
        ax.set_title("R=%g, r=%g\n\n" % (R, r) + r"$K_{SEC} = (\frac{R - r}{R})^3 = (1 - \frac{r}{R})^3 = (1 - \rho)^3$ = %.3g" % Ksec, y=0.95)
        ax.set_axis_off()

        x, y, z = sphere_xyz(center=(0.5, 0.5, 0.5), r=R, n=n, v=phi)
        ax.plot_surface(x, z, y, alpha=0.2, label="Pore: Radius=R=%.3g" % R)

        x, y, z = sphere_xyz(center=(0.5, 0.5, 0.5), r=R - r, n=n, v=phi)
        ax.plot_surface(x, z, y, alpha=0.2, label="Particle Center Boundary: Radius=R - r=%.3g" % (R - r))

        x, y, z = sphere_xyz(center=(0.5, r, 0.5), r=r)
        ax.plot_surface(x, z, y, facecolor='C2', alpha=0.2, label="Particles: Radius=r=%.3g" % r)

        x, y, z = sphere_xyz(center=(r, 0.5, 0.5), r=r)
        ax.plot_surface(x, z, y, facecolor='C2', alpha=0.2)

        x, y, z = sphere_xyz(center=(0.5, 1-r, 0.5), r=r)
        ax.plot_surface(x, z, y, facecolor='C2', alpha=0.2)

        ax.legend(loc="lower right")
        ax.view_init(elev=10, azim=-70, roll=0)

    fig.tight_layout()
    plt.show()

def illust_exclusion_coefficient():

    rhov = np.linspace(0.0, 1.0, 100)
    kv2 = np.power(1 - rhov, 2)
    kv3 = np.power(1 - rhov, 3)
    thv = np.arccos(1/(1 + rhov))
    kv_inv_4cyl = (1 - np.sqrt((2*rhov - rhov**2)) - (1-rhov)**2*(np.pi/4 - thv))/(1 - np.pi/4) 

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
    ax2.set_yscale('log')

    for ax in [ax1, ax2]:
        ax.plot(kv2, rhov)
        ax.plot(kv3, rhov)
        ax.plot(kv_inv_4cyl, rhov)

    fig.tight_layout()
    plt.show()


if __name__ == "__main__":
    import sys
    # import seaborn as sns
    # sns.set_theme()
    sys.path.append("../lib")

    # illust_cylindrical()
    # test_sphere_xyz_alt()
    # illust_spherical()
    illust_exclusion_coefficient()


"""
    Selective.VariedPropAnalysis.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import re
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar, get_color
from molass_legacy.KekLib.ScrolledFrame import ScrolledFrame
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

def get_code_text(editor):
    text = editor.advanced_frame.code_text.get(1.0, Tk.END)
    print("text=", text)
    return text

def check_code_text(code_text):
    context = {}
    exec(code_text, context)
    return True, context['proportions']

class VariedPropAnalysis(Dialog):
    def __init__(self, parent, editor, params_info, num_variations=20):
        self.parent = parent
        self.editor = editor
        code_text = get_code_text(editor)
        ret, func = check_code_text(code_text)
        self.proportions_func = func
        self.modelname = editor.get_current_modelname()
        self.params_info = params_info
        self.num_variations = num_variations
        self.current_index = None
        Dialog.__init__(self, parent, "Varied Proportion Analysis", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        self.scrolled_frame = ScrolledFrame(body_frame)
        self.scrolled_frame.pack(anchor=Tk.N)
        cframe = self.scrolled_frame.interior

        self.ncols = 5
        self.nrows = self.num_variations//self.ncols
        height = 2.5 * self.nrows
        fig, axes = plt.subplots(nrows=self.nrows, ncols=self.ncols, figsize=(20,height))
        self.draw_varied_proportions(fig, axes)
        fig.tight_layout()

        self.fig = fig
        self.axes = axes
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.popup_menu = None
        self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def buttonbox(self):
        # task: pack the close button so that it won't hide
        box = Tk.Frame(self)
        box.pack(fill=Tk.X)

        tframe = Tk.Frame(box)
        tframe.pack(side=Tk.LEFT, padx=20, pady=10)
        bframe = Tk.Frame(box)
        bframe.pack(side=Tk.RIGHT, padx=20, pady=10)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        w = Tk.Button(bframe, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=20)

        w = Tk.Button(bframe, text="OK", width=10, command=self.ok)
        w.pack(side=Tk.LEFT, padx=20)

        w = Tk.Button(bframe, text="RDR Chart", width=10, command=self.show_rdr_chart)
        w.pack(side=Tk.LEFT, padx=20)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)
        self.set_geometry()

    def set_geometry(self):
        self.update()
        canvas_width = int(self.mpl_canvas_widget.cget( 'width' ))
        canvas_height = int(self.mpl_canvas_widget.cget( 'height' ))
        print("canvas_height=", canvas_height)
        height = min(900, canvas_height+200)
        margin_width = 40
        wxh = '%dx%d' % (canvas_width + margin_width, height)
        geometry = self.geometry()
        new_geometry = re.sub( r'(\d+x\d+)(.+)', lambda m: wxh + m.group(2), geometry)
        self.geometry(new_geometry)

    def draw_varied_proportions(self, fig, axes, devel=True):
        if devel:
            from importlib import reload
            import Selective.PropOptimizerImpl
            reload(Selective.PropOptimizerImpl)
            import Selective.PropOptimizer
            reload(Selective.PropOptimizer)
        from Selective.PropOptimizerImpl import compute_range_rgs
        from Selective.PropOptimizer import PropOptimizer
        from molass_legacy._MOLASS.SerialSettings import set_setting
        from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve

        sd = self.editor.corbase_info.sd
        D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
        x = xr_curve.x
        y = xr_curve.y

        self.prop_optimizer = prop_optimizer = PropOptimizer(self.modelname, x, y)

        editor_ranges = self.editor.get_current_frame().editor_ranges

        # set_setting('local_debug', True)
        # ecurve = ElutionCurve(y, x=x)     # there still is a bug
        ecurve = ElutionCurve(y)
        self.paired_ranges = paired_ranges = ecurve.get_default_paired_ranges()
        # set_setting('local_debug', False)
        print("paired_ranges=", paired_ranges)
        print("editor_ranges=", editor_ranges)

        fig.suptitle("Varied Proportion Avalysis of %s using %s" % (get_in_folder(), self.modelname), fontsize=20)

        pv, fv_list, rgs_list, rdr_list, peaks_list = self.params_info

        self.min_n = n = np.argmin(rdr_list)
        minRDR = rdr_list[n]

        self.pv = pv

        for i, p in enumerate(self.pv):
            if i != 17:
                # continue
                pass

            print([i], "optimizing with p=%.3g" % p)

            opt_peaks = peaks_list[i]
            rdr = rdr_list[i]
            rgs = rgs_list[i]

            j, k = divmod(i,5)

            ax = axes[j,k]
            ax.plot(x, y, color="orange")

            cy_list = prop_optimizer.compute_cy_list(opt_peaks)
            for cy in cy_list:
                ax.plot(x, cy, ":")
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red")
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = xmin*0.9 + xmax*0.1
            ty = ymin*0.2 + ymax*0.8
            ax.text(tx, ty, "p=%.3g" % p, fontsize=20, alpha=0.3)
            ty = ymin*0.4 + ymax*0.6
            ax.text(tx, ty, "Rgs=%.3g, %.3g" % tuple(rgs), fontsize=20, alpha=0.3)

        if n is not None:
            j, k = divmod(n,5)
            ax = axes[j,k]
            ax.patch.set_facecolor('green')
            ax.patch.set_alpha(0.1)

        self.fv = np.array(fv_list)
        self.rdr = np.array(rdr_list)
        self.peaks_list = peaks_list

    def show_fv_chart(self):
        indeces = [str(i) for i in range(self.num_variations)]
        with dplt.Dp(window_title='FV Chart', ok_only=True, ok_text="Close"):
            fig, ax = dplt.subplots()
            ax.set_title("FV Chart", fontsize=16)
            ax.bar(indeces, self.fv)
            fig.tight_layout()
            dplt.show()

    def show_rdr_chart(self, devel=True):
        if devel:
            from importlib import reload
            import Selective.RdrChart
            reload(Selective.RdrChart)
        from .RdrChart import draw_rdr_chart
        title = "RDR_AD Chart of %s with %s" % (get_in_folder(), self.modelname)
        draw_rdr_chart(title, self.pv, self.rdr)

    def on_figure_click(self, event):
        if event.button == 3:
            axes = self.axes
            for i in range(axes.shape[0]):
                for j in range(axes.shape[1]):
                    if event.inaxes == axes[i,j]:
                        self.current_index = (i,j)
                        self.current_event = event
                        self.show_popup_menu(event)
                        break

    def show_popup_menu(self, event):
        from molass_legacy.KekLib.TkUtils import split_geometry
        self.create_popup_menu(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu(self, event):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0 )
            self.popup_menu.add_command(label='Copy Proportions', command=self.copy_proportions)

    def copy_proportions(self):
        k = self.get_current_seq_index()
        print([k], self.pv[k])

    def get_current_seq_index(self):
        if self.current_index is None:
            return self.min_n
        else:
            i, j = self.current_index
            return self.ncols*i + j

    def get_current_peaks(self):
        k = self.get_current_seq_index()
        return self.peaks_list[k]

    def validate(self):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        ret = MessageBox.askokcancel("Confirmation",
                "This will replace the decomposition with the optimal result.\n"
                + "Are you sure to proceed?",
                parent=self.parent)
        if ret:
            return 1
        else:
            return 0

    def apply(self):
        decomp_result = self.make_decomp_result()
        def apply_closure():
            self.editor.advanced_frameupdate_button_status(change_id="VPA")
            self.editor.update_current_frame_with_params(decomp_result)
        self.editor.after(100, apply_closure)

    def make_decomp_result(self, devel=True):
        if devel:
            from importlib import reload
            import Selective.V1ParamsAdapter
            reload(Selective.V1ParamsAdapter)
        from Selective.V1ParamsAdapter import make_decomp_result_impl

        peaks = self.get_current_peaks()
        return make_decomp_result_impl(self.editor, peaks)

def compute_peak_params(queue, modelname, sd, pv, fv_list, rgs_list, rdr_list, peaks_list, prop_func, devel=True):
    from time import sleep
    if devel:
        from importlib import reload
        import Selective.PropOptimizer
        reload(Selective.PropOptimizer)
        import Selective.PropOptimizer
        reload(Selective.PropOptimizer)
    from Selective.PropOptimizerImpl import compute_range_rgs
    from Selective.PropOptimizer import PropOptimizer

    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    x = xr_curve.x
    y = xr_curve.y
    prop_optimizer = PropOptimizer(modelname, x, y)
    init_peaks = prop_optimizer.get_init_params()
    props = prop_optimizer.compute_props(init_peaks)
    print("props=", props)
    paired_ranges = xr_curve.get_default_paired_ranges()

    for i, p in enumerate(pv):
        print([i], "optimizing with p=%.3g" % p)
        sleep(1)

        props = prop_func(p)
        opt_ret = prop_optimizer.optimize(props, init_params=init_peaks)
        opt_peaks = opt_ret.x.reshape(init_peaks.shape)
        cy_list = prop_optimizer.compute_cy_list(opt_peaks)
        C = np.array(cy_list)
        rgs = compute_range_rgs(qv, D, E, paired_ranges, C)
        rdr = abs(rgs[0] - rgs[1])*2/(rgs[0] + rgs[1])

        fv_list.append(opt_ret.fun)
        peaks_list.append(opt_peaks)
        rgs_list.append(rgs)
        rdr_list.append(rdr)
        if queue is None:
            print([i], "done")
        else:
            queue.put([i])

    if queue is None:
        print([-1], "finished")
    else:
        queue.put([-1])

def show_vp_analysis_impl(button_frame, prop_func=None, in_another_thread=True):
    from time import sleep
    if in_another_thread:
        import queue
        from threading import Thread

    editor = button_frame.editor
    modelname = editor.get_current_modelname()
    sd = editor.corbase_info.sd

    pv = button_frame.get_prop_vector()
    fv_list = []
    rgs_list = []
    rdr_list = []
    peaks_list = []

    if in_another_thread:
        queue = queue.Queue()
        thread = Thread(
                        target=compute_peak_params,
                        name='ComputePeakParams',
                        args=[queue, modelname, sd, pv, fv_list, rgs_list, rdr_list, peaks_list, prop_func]
                        )
        thread.start()

        while True:
            sleep(0.5)
            info = queue.get()
            if info is not None:
                button_frame.progress_update(info[0])
                if info[0] < 0:
                    break

        thread.join()
    else:
        compute_peak_params(None, modelname, sd, pv, fv_list, rgs_list, rdr_list, peaks_list)

    params_info = [pv, fv_list, rgs_list, rdr_list, peaks_list]
    dialog = VariedPropAnalysis(editor.parent, editor, params_info)
    dialog.show()



# coding: utf-8
"""
    AlmergeAnalyzer.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""

import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class AlmergeAnalyzerDialog( Dialog ):
    def __init__( self, parent, mc_vector, sd, ranges ):
        self.grab = 'local'     # used in grab_set
        self.parent     = parent
        self.mc_vector  = mc_vector
        self.sd         = sd
        self.mc_vector  = sd.mc_vector
        self.ranges     = ranges
        self.applied    = False

    def show( self ):
        title   = "Zero Concentration Extrapolation Solver"
        Dialog.__init__( self, self.parent, title )

    def body( self, body_fram ):
        tk_set_icon_portable( self )

        fframe = Tk.Frame( body_fram )
        fframe.pack()
        tframe = Tk.Frame( body_fram )
        tframe.pack( fill=Tk.X )

        cframe = Tk.Frame( fframe )
        cframe.pack( side=Tk.LEFT )
        pframe = Tk.Frame( fframe )
        pframe.pack( side=Tk.LEFT, fill=Tk.BOTH, padx=10 )

        self.fig = fig = plt.figure( figsize=(16, 8) )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.mpl_canvas.draw()
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        self.mpl_canvas.draw()

    def apply( self ):  # overrides parent class method
        self.applied = True

        zx_penalty_matrix = []
        for panel in self.panel_list:
            zx_penalty_matrix.append( panel.get_penalty_weights() )

        set_setting( 'zx_penalty_matrix', zx_penalty_matrix )



"""
    AffineContinuityDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.Affine import Affine
from .ElutionCurveModels import egha, emga

def demo():
    x = np.arange(300)
    mu = 150
    sigma = 30
    tau = 5
    a = 0

    try:
        np.linalg.inv(np.array([[1, 1], [1, 1]]))
    except ValueError:
        print("caught ValueError")
    except Exception:
        from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
        etb = ExceptionTracebacker()
        print(etb)

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,10))

        ax1, ax2 = axes[0,:]
        for a in np.linspace(-5, 5, 10):
            ax1.plot(x, egha(x, 1, mu, sigma, tau, a))
            ax2.plot(x, emga(x, 1, mu, sigma, tau, a))

        ax3, ax4 = axes[1,:]
        for a in np.linspace(-1e-7, 1e-7, 10):
            ax3.plot(x, egha(x, 1, mu, sigma, tau, a))
            ax4.plot(x, emga(x, 1, mu, sigma, tau, a))

        fig.tight_layout()
        plt.show()
